# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import eventlet
import math
import netaddr
import threading

from keystoneclient.auth.identity import v3
from keystoneclient import exceptions as k_exceptions
from keystoneclient import session
from keystoneclient.v2_0 import client as k_client
from keystoneclient.v3 import client
from neutron.common import utils
from neutron_lib import exceptions as n_exc
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import importutils
from oslo_utils import timeutils
from oslo_utils import uuidutils
from sqlalchemy import func
from sqlalchemy.orm import exc
from sqlalchemy.orm import joinedload
from sqlalchemy.sql import expression as expr

from networking_cisco._i18n import _
from networking_cisco import backwards_compatibility as bc
from networking_cisco.plugins.cisco.common import (cisco_constants as
                                                   c_constants)
from networking_cisco.plugins.cisco.db.device_manager import hd_models
from networking_cisco.plugins.cisco.db.device_manager import hosting_devices_db
from networking_cisco.plugins.cisco.device_manager import config
from networking_cisco.plugins.cisco.device_manager import service_vm_lib
from networking_cisco.plugins.cisco.extensions import ciscohostingdevicemanager

LOG = logging.getLogger(__name__)


HOSTING_DEVICE_MANAGER_OPTS = [
    cfg.StrOpt('l3_admin_tenant', default='L3AdminTenant',
               help=_("Name of the L3 admin tenant")),
    cfg.StrOpt('management_network', default='osn_mgmt_nw',
               help=_("Name of management network for service VM "
                      "configuration. Default value is osn_mgmt_nw")),
    cfg.StrOpt('default_security_group', default='mgmt_sec_grp',
               help=_("Default security group applied on management port. "
                      "Default value is mgmt_sec_grp")),
    cfg.BoolOpt('ensure_nova_running', default=True,
                help=_("Ensure that Nova is running before attempting to "
                       "create any service VM.")),
    cfg.StrOpt('domain_name_server_1', default='8.8.8.8',
               help=_("IP address of primary domain name server for hosting "
                      "devices")),
    cfg.StrOpt('domain_name_server_2', default='8.8.4.4',
               help=_("IP address of secondary domain name server for hosting "
                      "devices"))
]

cfg.CONF.register_opts(HOSTING_DEVICE_MANAGER_OPTS, "general")


VM_CATEGORY = ciscohostingdevicemanager.VM_CATEGORY


class HostingDeviceManagerMixin(hosting_devices_db.HostingDeviceDBMixin):
    """A class implementing a resource manager for hosting devices.

    The caller should make sure that HostingDeviceManagerMixin is a singleton.
    """

    # The all-mighty tenant owning all hosting devices
    _l3_tenant_uuid = None

    # Keystone session corresponding to admin user and l3_admin_tenant
    _keystone_session = None

    # The management network for hosting devices
    _mgmt_nw_uuid = None
    _mgmt_subnet_uuid = None
    _mgmt_sec_grp_id = None

    # Dictionary with credentials keyed on credential UUID
    _credentials = {}

    # Dictionaries with loaded driver modules for different host types
    _plugging_drivers = {}
    _hosting_device_drivers = {}

    # Dictionary with locks for hosting device pool maintenance
    _hosting_device_locks = {}

    # Service VM manager object that interacts with Nova
    _svc_vm_mgr_obj = None

    # Flag indicating is needed Nova services are reported as up.
    _nova_running = False

    @classmethod
    def _keystone_auth_session(cls):
        if cls._keystone_session:
            return cls._keystone_session
        else:
            auth_url = cfg.CONF.keystone_authtoken.auth_url
            if auth_url.rsplit('/', 1)[-1] != 'v3':
                auth_url += '/v3'
            # user = cfg.CONF.keystone_authtoken.admin_user
            # pw = cfg.CONF.keystone_authtoken.admin_password
            # project_name = cfg.CONF.keystone_authtoken.admin_tenant_name
            # project_name = cfg.CONF.keystone_authtoken.project_name
            user = cfg.CONF.keystone_authtoken.username
            pw = cfg.CONF.keystone_authtoken.password
            project_name = cfg.CONF.general.l3_admin_tenant
            user_domain_id = (cfg.CONF.keystone_authtoken.user_domain_id or
                              'default')
            project_domain_id = (cfg.CONF.keystone_authtoken.project_domain_id
                                 or 'default')
            auth = v3.Password(auth_url=auth_url,
                               username=user,
                               password=pw,
                               project_name=project_name,
                               user_domain_id=user_domain_id,
                               project_domain_id=project_domain_id)
            cls._keystone_session = session.Session(auth=auth)
        return cls._keystone_session

    @property
    def svc_vm_mgr(self):
        if self._svc_vm_mgr_obj is None:
            if hasattr(cfg.CONF.keystone_authtoken, 'project_domain_id'):
                self._svc_vm_mgr_obj = service_vm_lib.ServiceVMManager(
                    is_auth_v3=True,
                    keystone_session=self._keystone_auth_session())
            else:
                auth_url = cfg.CONF.keystone_authtoken.identity_uri + "/v2.0"
                u_name = cfg.CONF.keystone_authtoken.admin_user
                pw = cfg.CONF.keystone_authtoken.admin_password
                tenant = cfg.CONF.general.l3_admin_tenant

                self._svc_vm_mgr_obj = service_vm_lib.ServiceVMManager(
                    is_auth_v3=False,
                    user=u_name, passwd=pw, l3_admin_tenant=tenant,
                    auth_url=auth_url)
        return self._svc_vm_mgr_obj

    @classmethod
    def _get_tenant_id_using_keystone_v2(cls):
        auth_url = cfg.CONF.keystone_authtoken.identity_uri + "/v2.0"
        user = cfg.CONF.keystone_authtoken.admin_user
        pw = cfg.CONF.keystone_authtoken.admin_password
        tenant = cfg.CONF.keystone_authtoken.admin_tenant_name
        keystone = k_client.Client(username=user, password=pw,
                                   tenant_name=tenant,
                                   auth_url=auth_url)
        try:
            tenant = keystone.tenants.find(
                name=cfg.CONF.general.l3_admin_tenant)
        except k_exceptions.NotFound:
            LOG.error('No tenant with a name or ID of %s exists.',
                      cfg.CONF.general.l3_admin_tenant)
        except k_exceptions.NoUniqueMatch:
            LOG.error('Multiple tenants matches found for %s',
                      cfg.CONF.general.l3_admin_tenant)
        return tenant.id

    @classmethod
    def _get_tenant_id_using_keystone_v3(cls):
        keystone = client.Client(session=cls._keystone_auth_session())
        try:
            tenant = keystone.projects.find(
                name=cfg.CONF.general.l3_admin_tenant)
        except k_exceptions.NotFound:
            LOG.error('No tenant with a name or ID of %s exists.',
                      cfg.CONF.general.l3_admin_tenant)
        except k_exceptions.NoUniqueMatch:
            LOG.error('Multiple tenants matches found for %s',
                      cfg.CONF.general.l3_admin_tenant)
        return tenant.id

    @classmethod
    def l3_tenant_id(cls):
        """Returns id of tenant owning hosting device resources."""
        if cls._l3_tenant_uuid is None:
            if hasattr(cfg.CONF.keystone_authtoken, 'project_domain_id'):
                # TODO(sridar): hack for now to determing if keystone v3
                # API is to be used.
                cls._l3_tenant_uuid = cls._get_tenant_id_using_keystone_v3()
            else:
                cls._l3_tenant_uuid = cls._get_tenant_id_using_keystone_v2()
        return cls._l3_tenant_uuid

    @classmethod
    def mgmt_nw_id(cls):
        """Returns id of the management network."""
        if cls._mgmt_nw_uuid is None:
            tenant_id = cls.l3_tenant_id()
            if not tenant_id:
                return
            net = bc.get_plugin().get_networks(
                bc.context.get_admin_context(),
                {'tenant_id': [tenant_id],
                 'name': [cfg.CONF.general.management_network]},
                ['id', 'subnets'])
            if len(net) == 1:
                num_subnets = len(net[0]['subnets'])
                if num_subnets == 0:
                    LOG.error('The management network has no subnet. '
                              'Please assign one.')
                    return
                elif num_subnets > 1:
                    LOG.info('The management network has %d subnets. The '
                             'first one will be used.', num_subnets)
                cls._mgmt_nw_uuid = net[0].get('id')
                cls._mgmt_subnet_uuid = net[0]['subnets'][0]
            elif len(net) > 1:
                # Management network must have a unique name.
                LOG.error('The management network for does not have '
                          'unique name. Please ensure that it is.')
            else:
                # Management network has not been created.
                LOG.error('There is no virtual management network. Please '
                          'create one.')
        return cls._mgmt_nw_uuid

    @classmethod
    def mgmt_subnet_id(cls):
        if cls._mgmt_subnet_uuid is None:
            cls.mgmt_nw_id()
        return cls._mgmt_subnet_uuid

    @classmethod
    def mgmt_sec_grp_id(cls):
        """Returns id of security group used by the management network."""
        if not utils.is_extension_supported(bc.get_plugin(), "security-group"):
            return
        if cls._mgmt_sec_grp_id is None:
            # Get the id for the _mgmt_security_group_id
            tenant_id = cls.l3_tenant_id()
            res = bc.get_plugin().get_security_groups(
                bc.context.get_admin_context(),
                {'tenant_id': [tenant_id],
                 'name': [cfg.CONF.general.default_security_group]},
                ['id'])
            if len(res) == 1:
                sec_grp_id = res[0].get('id', None)
                cls._mgmt_sec_grp_id = sec_grp_id
            elif len(res) > 1:
                # the mgmt sec group must be unique.
                LOG.error('The security group for the management network '
                          'does not have unique name. Please ensure that '
                          'it is.')
            else:
                # Service VM Mgmt security group is not present.
                LOG.error('There is no security group for the management '
                          'network. Please create one.')
        return cls._mgmt_sec_grp_id

    def get_hosting_device_config(self, context, id):
        # ask config agent for the running config of the hosting device
        cfg_notifier = self.agent_notifiers.get(c_constants.AGENT_TYPE_CFG)
        if cfg_notifier:
            return cfg_notifier.get_hosting_device_configuration(context, id)

    def get_hosting_device_driver(self, context, id):
        """Returns device driver for hosting device template with <id>."""
        if id is None:
            return
        try:
            return self._hosting_device_drivers[id]
        except KeyError:
            try:
                template = self._get_hosting_device_template(context, id)
                self._hosting_device_drivers[id] = importutils.import_object(
                    template['device_driver'])
            except (ImportError, TypeError, n_exc.NeutronException):
                LOG.exception("Error loading hosting device driver for "
                              "hosting device template %s", id)
            return self._hosting_device_drivers.get(id)

    def get_hosting_device_plugging_driver(self, context, id):
        """Returns  plugging driver for hosting device template with <id>."""
        if id is None:
            return
        try:
            return self._plugging_drivers[id]
        except KeyError:
            try:
                template = self._get_hosting_device_template(context, id)
                self._plugging_drivers[id] = importutils.import_object(
                    template['plugging_driver'])
            except (ImportError, TypeError, n_exc.NeutronException):
                LOG.exception("Error loading plugging driver for hosting "
                              "device template %s", id)
            return self._plugging_drivers.get(id)

    def report_hosting_device_shortage(self, context, template, requested=0):
        """Used to report shortage of hosting devices based on <template>."""
        self._dispatch_pool_maintenance_job(template)

    def acquire_hosting_device_slots(self, context, hosting_device, resource,
                                     resource_type, resource_service, num,
                                     exclusive=False):
        """Assign <num> slots in <hosting_device> to logical <resource>.

        If exclusive is True the hosting device is bound to the resource's
        tenant. Otherwise it is not bound to any tenant.

        Returns True if allocation was granted, False otherwise.
        """
        bound = hosting_device['tenant_bound']
        if ((bound is not None and bound != resource['tenant_id']) or
            (exclusive and not self._exclusively_used(context, hosting_device,
                                                      resource['tenant_id']))):
            LOG.debug(
                'Rejecting allocation of %(num)d slots in tenant %(bound)s '
                'hosting device %(device)s to logical resource %(r_id)s due '
                'to exclusive use conflict.',
                {'num': num,
                 'bound': 'unbound' if bound is None else bound + ' bound',
                 'device': hosting_device['id'], 'r_id': resource['id']})
            return False
        with context.session.begin(subtransactions=True):
            res_info = {'resource': resource, 'type': resource_type,
                        'service': resource_service}
            slot_info, query = self._get_or_create_slot_allocation(
                context, hosting_device, res_info)
            if slot_info is None:
                LOG.debug('Rejecting allocation of %(num)d slots in hosting '
                          'device %(device)s to logical resource %(r_id)s',
                          {'num': num, 'device': hosting_device['id'],
                           'r_id': resource['id']})
                return False
            new_allocation = num + slot_info.num_allocated
            if hosting_device['template']['slot_capacity'] < new_allocation:
                LOG.debug('Rejecting allocation of %(num)d slots in '
                          'hosting device %(device)s to logical resource '
                          '%(r_id)s due to insufficent slot availability.',
                          {'num': num, 'device': hosting_device['id'],
                           'r_id': resource['id']})
                self._dispatch_pool_maintenance_job(hosting_device['template'])
                return False
            # handle any changes to exclusive usage by tenant
            if exclusive and bound is None:
                self._update_hosting_device_exclusivity(
                    context, hosting_device, resource['tenant_id'])
                bound = resource['tenant_id']
            elif not exclusive and bound is not None:
                self._update_hosting_device_exclusivity(context,
                                                        hosting_device, None)
                bound = None
            slot_info.num_allocated = new_allocation
            context.session.add(slot_info)
        self._dispatch_pool_maintenance_job(hosting_device['template'])
        # report success
        LOG.info('Allocated %(num)d additional slots in tenant %(bound)s'
                 'bound hosting device %(hd_id)s. In total %(total)d '
                 'slots are now allocated in that hosting device for '
                 'logical resource %(r_id)s.',
                 {'num': num, 'bound': 'un-' if bound is None else bound + ' ',
                  'total': new_allocation, 'hd_id': hosting_device['id'],
                  'r_id': resource['id']})
        return True

    def release_hosting_device_slots(self, context, hosting_device, resource,
                                     num):
        """Free <num> slots in <hosting_device> from logical resource <id>.

        Returns True if deallocation was successful. False otherwise.
        """
        with context.session.begin(subtransactions=True):
            num_str = str(num) if num >= 0 else "all"
            res_info = {'resource': resource}
            slot_info, query = self._get_or_create_slot_allocation(
                context, hosting_device, res_info, create=False)
            if slot_info is None:
                LOG.debug('Rejecting de-allocation of %(num)s slots in '
                          'hosting device %(device)s for logical resource '
                          '%(id)s', {'num': num_str,
                                     'device': hosting_device['id'],
                                     'id': resource['id']})
                return False
            if num >= 0:
                new_allocation = slot_info.num_allocated - num
            else:
                # if a negative num is specified all slot allocations for
                # the logical resource in the hosting device is removed
                new_allocation = 0
            if new_allocation < 0:
                LOG.debug('Rejecting de-allocation of %(num)s slots in '
                          'hosting device %(device)s for logical resource '
                          '%(id)s since only %(alloc)d slots are allocated.',
                          {'num': num_str, 'device': hosting_device['id'],
                           'id': resource['id'],
                           'alloc': slot_info.num_allocated})
                self._dispatch_pool_maintenance_job(hosting_device['template'])
                return False
            elif new_allocation == 0:
                result = query.delete()
                LOG.info('De-allocated %(num)s slots from hosting device '
                         '%(hd_id)s. %(total)d slots are now allocated in '
                         'that hosting device.',
                         {'num': num_str, 'total': new_allocation,
                          'hd_id': hosting_device['id']})
                if (hosting_device['tenant_bound'] is not None and
                    context.session.query(hd_models.SlotAllocation).filter_by(
                        hosting_device_id=hosting_device['id']).first() is
                        None):
                    # make hosting device tenant unbound if no logical
                    # resource use it anymore
                    hosting_device['tenant_bound'] = None
                    context.session.add(hosting_device)
                    LOG.info('Making hosting device %(hd_id)s with no '
                             'allocated slots tenant unbound.',
                             {'hd_id': hosting_device['id']})
                self._dispatch_pool_maintenance_job(hosting_device['template'])
                return result == 1
            LOG.info('De-allocated %(num)s slots from hosting device '
                     '%(hd_id)s. %(total)d slots are now allocated in '
                     'that hosting device.',
                     {'num': num_str, 'total': new_allocation,
                      'hd_id': hosting_device['id']})
            slot_info.num_allocated = new_allocation
            context.session.add(slot_info)
        self._dispatch_pool_maintenance_job(hosting_device['template'])
        # report success
        return True

    def _get_or_create_slot_allocation(self, context, hosting_device,
                                       resource_info, create=True):
        resource = resource_info['resource']
        slot_info = None
        query = context.session.query(hd_models.SlotAllocation).filter_by(
            logical_resource_id=resource['id'],
            hosting_device_id=hosting_device['id'])
        with context.session.begin(subtransactions=True):
            try:
                slot_info = query.one()
            except exc.MultipleResultsFound:
                # this should not happen
                LOG.debug('DB inconsistency: Multiple slot allocation entries '
                          'for logical resource %(r_id)s in hosting device '
                          '%(device)s.', {'r_id': resource['id'],
                                          'device': hosting_device['id']})
            except exc.NoResultFound:
                LOG.debug('Logical resource %(res)s does not have allocated '
                          'any slots in hosting device %(dev)s.',
                          {'res': resource['id'], 'dev': hosting_device['id']})
                if create is True:
                    LOG.debug('Creating new slot allocation DB entry for '
                              'logical resource %(res)s in hosting device '
                              '%(dev)s.', {'res': resource['id'],
                                           'dev': hosting_device['id']})
                    slot_info = hd_models.SlotAllocation(
                        template_id=hosting_device['template_id'],
                        hosting_device_id=hosting_device['id'],
                        logical_resource_type=resource_info['type'],
                        logical_resource_service=resource_info['service'],
                        logical_resource_id=resource['id'],
                        logical_resource_owner=resource['tenant_id'],
                        num_allocated=0,
                        tenant_bound=None)
        return slot_info, query

    def get_slot_allocation(self, context, template_id=None,
                            hosting_device_id=None, resource_id=None):
        query = context.session.query(func.sum(
            hd_models.SlotAllocation.num_allocated))
        if template_id is not None:
            query = query.filter_by(template_id=template_id)
        if hosting_device_id is not None:
            query = query.filter_by(hosting_device_id=hosting_device_id)
        if resource_id is not None:
            query = query.filter_by(logical_resource_id=resource_id)
        return query.scalar() or 0

    def get_hosting_devices_qry(self, context, hosting_device_ids,
                                load_agent=True):
        """Returns hosting devices with <hosting_device_ids>."""
        query = context.session.query(hd_models.HostingDevice)
        if load_agent:
            query = query.options(joinedload('cfg_agent'))
        if len(hosting_device_ids) > 1:
            query = query.filter(hd_models.HostingDevice.id.in_(
                hosting_device_ids))
        else:
            query = query.filter(hd_models.HostingDevice.id ==
                                 hosting_device_ids[0])
        return query

    def delete_all_hosting_devices(self, context, force_delete=False):
        """Deletes all hosting devices."""
        for item in self._get_collection_query(
                context, hd_models.HostingDeviceTemplate):
            self.delete_all_hosting_devices_by_template(
                context, template=item, force_delete=force_delete)

    def delete_all_hosting_devices_by_template(self, context, template,
                                               force_delete=False):
        """Deletes all hosting devices based on <template>."""
        plugging_drv = self.get_hosting_device_plugging_driver(
            context, template['id'])
        hosting_device_drv = self.get_hosting_device_driver(context,
                                                            template['id'])
        if plugging_drv is None or hosting_device_drv is None:
            return
        is_vm = template['host_category'] == VM_CATEGORY
        query = context.session.query(hd_models.HostingDevice)
        query = query.filter(hd_models.HostingDevice.template_id ==
                             template['id'])
        for hd in query:
            if not (hd.auto_delete or force_delete):
                # device manager is not responsible for life cycle
                # management of this hosting device.
                continue
            res = plugging_drv.get_hosting_device_resources(
                context, hd.id, hd.complementary_id, self.l3_tenant_id(),
                self.mgmt_nw_id())
            if is_vm:
                self.svc_vm_mgr.delete_service_vm(context, hd.id)
            plugging_drv.delete_hosting_device_resources(
                context, self.l3_tenant_id(), **res)
            with context.session.begin(subtransactions=True):
                # remove all allocations in this hosting device
                context.session.query(hd_models.SlotAllocation).filter_by(
                    hosting_device_id=hd['id']).delete()
                context.session.delete(hd)

    def handle_non_responding_hosting_devices(self, context, cfg_agent,
                                              hosting_device_ids):
        e_context = context.elevated()
        hosting_devices = self.get_hosting_devices_qry(
            e_context, hosting_device_ids).all()
        # 'hosting_info' is dictionary with ids of removed hosting
        # devices and the affected logical resources for each
        # removed hosting device:
        #    {'hd_id1': {'routers': [id1, id2, ...],
        #                'fw': [id1, ...],
        #                 ...},
        #     'hd_id2': {'routers': [id3, id4, ...]},
        #                'fw': [id1, ...],
        #                ...},
        #     ...}
        hosting_info = dict((id, {}) for id in hosting_device_ids)
        #TODO(bobmel): Modify so service plugins register themselves
        try:
            l3plugin = bc.get_plugin(bc.constants.L3)
            l3plugin.handle_non_responding_hosting_devices(
                context, hosting_devices, hosting_info)
        except AttributeError:
            pass
        notifier = self.agent_notifiers.get(c_constants.AGENT_TYPE_CFG)
        for hd in hosting_devices:
            if (self._process_non_responsive_hosting_device(e_context, hd) and
                    notifier):
                notifier.hosting_devices_removed(context, hosting_info, False,
                                                 cfg_agent)

    def get_device_info_for_agent(self, context, hosting_device_db):
        """Returns information about <hosting_device> needed by config agent.

           Convenience function that service plugins can use to populate
           their resources with information about the device hosting their
           logical resource.
        """
        template = hosting_device_db.template
        mgmt_port = hosting_device_db.management_port
        mgmt_ip = (mgmt_port['fixed_ips'][0]['ip_address']
                   if mgmt_port else hosting_device_db.management_ip_address)
        return {'id': hosting_device_db.id,
                'name': template.name,
                'template_id': template.id,
                'credentials': self._get_credentials(hosting_device_db),
                'host_category': template.host_category,
                'admin_state_up': hosting_device_db.admin_state_up,
                'service_types': template.service_types,
                'management_ip_address': mgmt_ip,
                'protocol_port': hosting_device_db.protocol_port,
                'timeout': None,
                'created_at': str(hosting_device_db.created_at),
                'status': hosting_device_db.status,
                'booting_time': template.booting_time}

    def _process_non_responsive_hosting_device(self, context, hosting_device):
        """Host type specific processing of non responsive hosting devices.

        :param hosting_device: db object for hosting device
        :return: True if hosting_device has been deleted, otherwise False
        """
        if (hosting_device['template']['host_category'] == VM_CATEGORY and
                hosting_device['auto_delete']):
            self._delete_dead_service_vm_hosting_device(context,
                                                        hosting_device)
            return True
        return False

    def _setup_device_manager(self):
        self._credentials = (
            config.obtain_hosting_device_credentials_from_config())
        self._create_hosting_device_templates_from_config()
        self._create_hosting_devices_from_config()
        self._gt_pool = eventlet.GreenPool()
        # initialize hosting device pools
        adm_ctx = bc.context.get_admin_context()
        for template in adm_ctx.session.query(hd_models.HostingDeviceTemplate):
            self._dispatch_pool_maintenance_job(template)

    def _dispatch_pool_maintenance_job(self, template):
        # Note(bobmel): Nova does not handle VM dispatching well before all
        # its services have started. This creates problems for the Neutron
        # devstack script that creates a Neutron router, which in turn
        # triggers service VM dispatching.
        # Only perform pool maintenance if needed Nova services have started

        # For now the pool size is only elastic for service VMs.
        if template['host_category'] != VM_CATEGORY:
            return
        if cfg.CONF.general.ensure_nova_running and not self._nova_running:
            if self.svc_vm_mgr.nova_services_up():
                self._nova_running = True
            else:
                LOG.info('Not all Nova services are up and running. '
                         'Skipping this service vm pool management '
                         'request.')
                return
        adm_context = bc.context.get_admin_context()
        adm_context.tenant_id = self.l3_tenant_id()
        self._gt_pool.spawn_n(self._maintain_hosting_device_pool, adm_context,
                              template)

    def _maintain_hosting_device_pool(self, context, template):
        """Maintains the pool of hosting devices that are based on <template>.

        Ensures that the number of standby hosting devices (essentially
        service VMs) is kept at a suitable level so that resource creation is
        not slowed down by booting of the hosting device.

        :param context: context for this operation
        :param template: db object for hosting device template
        """
        #TODO(bobmel): Support HA/load-balanced Neutron servers:
        #TODO(bobmel): Locking across multiple running Neutron server instances
        lock = self._get_template_pool_lock(template['id'])
        acquired = lock.acquire(False)
        if not acquired:
            # pool maintenance for this template already ongoing, so abort
            return
        try:
            # Maintain a pool of approximately 'desired_slots_free' available
            # for allocation. Approximately means that
            # abs(desired_slots_free-capacity) <= available_slots <=
            #                                       desired_slots_free+capacity
            capacity = template['slot_capacity']
            if capacity == 0:
                return
            desired = template['desired_slots_free']
            available = self._get_total_available_slots(
                context, template['id'], capacity)
            grow_threshold = abs(desired - capacity)
            if available <= grow_threshold:
                num_req = int(math.ceil(grow_threshold / (1.0 * capacity)))
                num_created = len(self._create_svc_vm_hosting_devices(
                    context, num_req, template))
                if num_created < num_req:
                    LOG.warning('Requested %(requested)d instances based '
                                'on hosting device template %(template)s '
                                'but could only create %(created)d '
                                'instances',
                                {'requested': num_req,
                                 'template': template['id'],
                                 'created': num_created})
            elif available >= desired + capacity:
                num_req = int(
                    math.floor((available - desired) / (1.0 * capacity)))
                num_deleted = self._delete_idle_service_vm_hosting_devices(
                    context, num_req, template)
                if num_deleted < num_req:
                    LOG.warning('Tried to delete %(requested)d instances '
                                'based on hosting device template '
                                '%(template)s but could only delete '
                                '%(deleted)d instances',
                             {'requested': num_req, 'template': template['id'],
                              'deleted': num_deleted})
        finally:
            lock.release()

    def _create_svc_vm_hosting_devices(self, context, num, template):
        """Creates <num> or less service VM instances based on <template>.

        These hosting devices can be bound to a certain tenant or for shared
        use. A list with the created hosting device VMs is returned.
        """
        hosting_devices = []
        template_id = template['id']
        credentials_id = template['default_credentials_id']
        plugging_drv = self.get_hosting_device_plugging_driver(context,
                                                               template_id)
        hosting_device_drv = self.get_hosting_device_driver(context,
                                                            template_id)
        if plugging_drv is None or hosting_device_drv is None or num <= 0:
            return hosting_devices
        #TODO(bobmel): Determine value for max_hosted properly
        max_hosted = 1  # template['slot_capacity']
        dev_data, mgmt_context = self._get_resources_properties_for_hd(
            template, credentials_id)
        credentials_info = self._credentials.get(credentials_id)
        if credentials_info is None:
            LOG.error('Could not find credentials for hosting device'
                      'template %s. Aborting VM hosting device creation.',
                      template_id)
            return hosting_devices
        connectivity_info = self._get_mgmt_connectivity_info(
            context, self.mgmt_subnet_id())
        for i in range(num):
            complementary_id = uuidutils.generate_uuid()
            res = plugging_drv.create_hosting_device_resources(
                context, complementary_id, self.l3_tenant_id(), mgmt_context,
                max_hosted)
            if res.get('mgmt_port') is None:
                # Required ports could not be created
                return hosting_devices
            connectivity_info['mgmt_port'] = res['mgmt_port']
            vm_instance = self.svc_vm_mgr.dispatch_service_vm(
                context, template['name'] + '_nrouter', template['image'],
                template['flavor'], hosting_device_drv, credentials_info,
                connectivity_info, res.get('ports'))
            if vm_instance is not None:
                dev_data.update(
                    {'id': vm_instance['id'],
                     'complementary_id': complementary_id,
                     'management_ip_address': res['mgmt_port'][
                         'fixed_ips'][0]['ip_address'],
                     'management_port_id': res['mgmt_port']['id']})
                self.create_hosting_device(context,
                                           {'hosting_device': dev_data})
                hosting_devices.append(vm_instance)
            else:
                # Fundamental error like could not contact Nova
                # Cleanup anything we created
                plugging_drv.delete_hosting_device_resources(
                    context, self.l3_tenant_id(), **res)
                break
        LOG.info('Created %(num)d hosting device VMs based on template '
                 '%(t_id)s', {'num': len(hosting_devices),
                              't_id': template_id})
        return hosting_devices

    def _get_mgmt_connectivity_info(self, context, mgmt_subnet_id):
        subnet_data = self._core_plugin.get_subnet(
            context, mgmt_subnet_id,
            ['cidr', 'gateway_ip', 'dns_nameservers'])
        num = len(subnet_data['dns_nameservers'])
        name_server_1 = cfg.CONF.general.domain_name_server_1
        name_server_2 = cfg.CONF.general.domain_name_server_2
        if num == 1:
            name_server_1 = subnet_data['dns_nameservers'][0]['address']
            name_server_2 = cfg.CONF.general.domain_name_server_2
        elif num >= 2:
            name_server_1 = subnet_data['dns_nameservers'][0]['address']
            name_server_2 = subnet_data['dns_nameservers'][1]['address']
        return {'gateway_ip': subnet_data['gateway_ip'],
                'netmask': str(netaddr.IPNetwork(subnet_data['cidr']).netmask),
                'name_server_1': name_server_1,
                'name_server_2': name_server_2}

    def _get_resources_properties_for_hd(self, template, credentials_id):
        # These resources are owned by the L3AdminTenant
        dev_data = {'template_id': template['id'],
                    'tenant_id': template['tenant_id'],
                    'credentials_id': credentials_id,
                    'admin_state_up': True,
                    'protocol_port': template['protocol_port'],
                    'created_at': timeutils.utcnow(),
                    'tenant_bound': template['tenant_bound'] or None,
                    'auto_delete': True}
        mgmt_context = {
            'mgmt_ip_address': None,
            'mgmt_nw_id': self.mgmt_nw_id(),
            'mgmt_sec_grp_id': self.mgmt_sec_grp_id()}
        return dev_data, mgmt_context

    def _delete_idle_service_vm_hosting_devices(self, context, num, template):
        """Deletes <num> or less unused <template>-based service VM instances.

        The number of deleted service vm instances is returned.
        """
        # Delete the "youngest" hosting devices since they are more likely
        # not to have finished booting
        num_deleted = 0
        plugging_drv = self.get_hosting_device_plugging_driver(context,
                                                               template['id'])
        hosting_device_drv = self.get_hosting_device_driver(context,
                                                            template['id'])
        if plugging_drv is None or hosting_device_drv is None or num <= 0:
            return num_deleted
        query = context.session.query(hd_models.HostingDevice)
        query = query.outerjoin(
            hd_models.SlotAllocation,
            hd_models.HostingDevice.id ==
            hd_models.SlotAllocation.hosting_device_id)
        query = query.filter(hd_models.HostingDevice.template_id ==
                             template['id'],
                             hd_models.HostingDevice.admin_state_up ==
                             expr.true(),
                             hd_models.HostingDevice.tenant_bound ==
                             expr.null(),
                             hd_models.HostingDevice.auto_delete ==
                             expr.true())
        query = query.group_by(hd_models.HostingDevice.id).having(
            func.count(hd_models.SlotAllocation.logical_resource_id) == 0)
        query = query.order_by(
            hd_models.HostingDevice.created_at.desc(),
            func.count(hd_models.SlotAllocation.logical_resource_id))
        hd_candidates = query.all()
        num_possible_to_delete = min(len(hd_candidates), num)
        for i in range(num_possible_to_delete):
            res = plugging_drv.get_hosting_device_resources(
                context, hd_candidates[i]['id'],
                hd_candidates[i]['complementary_id'], self.l3_tenant_id(),
                self.mgmt_nw_id())
            if self.svc_vm_mgr.delete_service_vm(context,
                                                 hd_candidates[i]['id']):
                with context.session.begin(subtransactions=True):
                    context.session.delete(hd_candidates[i])
                plugging_drv.delete_hosting_device_resources(
                    context, self.l3_tenant_id(), **res)
                num_deleted += 1
        LOG.info('Deleted %(num)d hosting devices based on template '
                 '%(t_id)s', {'num': num_deleted, 't_id': template['id']})
        return num_deleted

    def _delete_dead_service_vm_hosting_device(self, context, hosting_device):
        """Deletes a presumably dead <hosting_device> service VM.

        This will indirectly make all of its hosted resources unscheduled.
        """
        if hosting_device is None:
            return
        plugging_drv = self.get_hosting_device_plugging_driver(
            context, hosting_device['template_id'])
        hosting_device_drv = self.get_hosting_device_driver(
            context, hosting_device['template_id'])
        if plugging_drv is None or hosting_device_drv is None:
            return
        res = plugging_drv.get_hosting_device_resources(
            context, hosting_device['id'], hosting_device['complementary_id'],
            self.l3_tenant_id(), self.mgmt_nw_id())
        if not self.svc_vm_mgr.delete_service_vm(context,
                                                 hosting_device['id']):
            LOG.error('Failed to delete hosting device %s service VM. '
                      'Will un-register it anyway.',
                      hosting_device['id'])
        plugging_drv.delete_hosting_device_resources(
            context, self.l3_tenant_id(), **res)
        with context.session.begin(subtransactions=True):
            # remove all allocations in this hosting device
            context.session.query(hd_models.SlotAllocation).filter_by(
                hosting_device_id=hosting_device['id']).delete()
            context.session.delete(hosting_device)

    def _get_total_available_slots(self, context, template_id, capacity):
        """Returns available slots in idle devices based on <template_id>.

        Only slots in tenant unbound hosting devices are counted to ensure
        there is always hosting device slots available regardless of tenant.
        """
        query = context.session.query(hd_models.HostingDevice.id)
        query = query.outerjoin(
            hd_models.SlotAllocation,
            hd_models.HostingDevice.id == hd_models.SlotAllocation
            .hosting_device_id)
        query = query.filter(
            hd_models.HostingDevice.template_id == template_id,
            hd_models.HostingDevice.admin_state_up == expr.true(),
            hd_models.HostingDevice.tenant_bound == expr.null())
        query = query.group_by(hd_models.HostingDevice.id)
        query = query.having(
            func.sum(hd_models.SlotAllocation.num_allocated) == expr.null())
        num_hosting_devices = query.count()
        return num_hosting_devices * capacity

    def _exclusively_used(self, context, hosting_device, tenant_id):
        """Checks if only <tenant_id>'s resources use <hosting_device>."""
        return (context.session.query(hd_models.SlotAllocation).filter(
            hd_models.SlotAllocation.hosting_device_id == hosting_device['id'],
            hd_models.SlotAllocation.logical_resource_owner != tenant_id).
            first() is None)

    def _update_hosting_device_exclusivity(self, context, hosting_device,
                                           tenant_id):
        """Make <hosting device> bound or unbound to <tenant_id>.

        If <tenant_id> is None the device is unbound, otherwise it gets bound
        to that <tenant_id>
        """
        with context.session.begin(subtransactions=True):
            hosting_device['tenant_bound'] = tenant_id
            context.session.add(hosting_device)
            for item in (context.session.query(hd_models.SlotAllocation).
                         filter_by(hosting_device_id=hosting_device['id'])):
                item['tenant_bound'] = tenant_id
                context.session.add(item)

    def _get_template_pool_lock(self, id):
        """Returns lock object for hosting device template with <id>."""
        try:
            return self._hosting_device_locks[id]
        except KeyError:
            self._hosting_device_locks[id] = threading.Lock()
            return self._hosting_device_locks.get(id)

    def _get_credentials(self, hosting_device_db, obfuscate=True):
        creds_id = (hosting_device_db.credentials_id
                    if hosting_device_db.credentials_id in self._credentials
                    else hosting_device_db.template.default_credentials_id)
        creds = self._credentials.get(creds_id)
        if creds is None:
            return None
        password = '********' if obfuscate is True else creds['password']
        return {'credentials_id': creds_id, 'user_name': creds['user_name'],
                'password': password}

    def _create_hosting_device_templates_from_config(self):
        """To be called late during plugin initialization so that any hosting
        device templates defined in the config file is properly inserted in
        the DB.
        """
        hdt_dict = config.get_specific_config('cisco_hosting_device_template')
        attr_info = ciscohostingdevicemanager.RESOURCE_ATTRIBUTE_MAP[
            ciscohostingdevicemanager.DEVICE_TEMPLATES]
        adm_context = bc.context.get_admin_context()

        for hdt_uuid, kv_dict in hdt_dict.items():
            # ensure hdt_uuid is properly formatted
            hdt_uuid = config.uuidify(hdt_uuid)
            try:
                self.get_hosting_device_template(adm_context, hdt_uuid)
                is_create = False
            except ciscohostingdevicemanager.HostingDeviceTemplateNotFound:
                is_create = True
            kv_dict['id'] = hdt_uuid
            kv_dict['tenant_id'] = self.l3_tenant_id()
            config.verify_resource_dict(kv_dict, True, attr_info)
            hdt = {ciscohostingdevicemanager.DEVICE_TEMPLATE: kv_dict}
            try:
                if is_create:
                    self.create_hosting_device_template(adm_context, hdt)
                else:
                    self.update_hosting_device_template(adm_context,
                                                        kv_dict['id'], hdt)
            except n_exc.NeutronException:
                with excutils.save_and_reraise_exception():
                    LOG.error('Invalid hosting device template definition '
                              'in configuration file for template = %s',
                              hdt_uuid)

    def _create_hosting_devices_from_config(self):
        """To be called late during plugin initialization so that any hosting
        device specified in the config file is properly inserted in the DB.
        """
        hd_dict = config.get_specific_config('cisco_hosting_device')
        attr_info = ciscohostingdevicemanager.RESOURCE_ATTRIBUTE_MAP[
            ciscohostingdevicemanager.DEVICES]
        adm_context = bc.context.get_admin_context()

        for hd_uuid, kv_dict in hd_dict.items():
            # ensure hd_uuid is properly formatted
            hd_uuid = config.uuidify(hd_uuid)
            try:
                old_hd = self.get_hosting_device(adm_context, hd_uuid)
                is_create = False
            except ciscohostingdevicemanager.HostingDeviceNotFound:
                old_hd = {}
                is_create = True
            kv_dict['id'] = hd_uuid
            kv_dict['tenant_id'] = self.l3_tenant_id()
            # make sure we keep using same config agent if it has been assigned
            kv_dict['cfg_agent_id'] = old_hd.get('cfg_agent_id')
            # make sure we keep using management port if it exists
            kv_dict['management_port_id'] = old_hd.get('management_port_id')
            config.verify_resource_dict(kv_dict, True, attr_info)
            hd = {ciscohostingdevicemanager.DEVICE: kv_dict}
            try:
                if is_create:
                    self.create_hosting_device(adm_context, hd)
                else:
                    self.update_hosting_device(adm_context, kv_dict['id'], hd)
            except n_exc.NeutronException:
                with excutils.save_and_reraise_exception():
                    LOG.error('Invalid hosting device specification in '
                              'configuration file for device = %s',
                              hd_uuid)



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from neutron.tests import base

from neutron_lib import exceptions as n_exc

from networking_cisco.plugins.cisco.device_manager.plugging_drivers.\
    vif_hotplug_plugging_driver import(VIFHotPlugPluggingDriver)


class TestVIFHotPlugPluggingDriver(base.BaseTestCase):

    def setUp(self):
        super(TestVIFHotPlugPluggingDriver, self).setUp()

    def test_delete_resource_port_fail_always(self):
        mgmt_port_id = 'fake_port_id'
        mocked_plugin = mock.MagicMock()
        mock_ctx = mock.MagicMock()
        mocked_plugin.delete_port = mock.MagicMock(
            side_effect=n_exc.NeutronException)

        with mock.patch.object(VIFHotPlugPluggingDriver, '_core_plugin') as (
                plugin):
            plugin.__get__ = mock.MagicMock(return_value=mocked_plugin)
            vif_plugging_driver = VIFHotPlugPluggingDriver()
            self.assertRaises(
                n_exc.NeutronException,
                vif_plugging_driver._delete_resource_port,
                mock_ctx,
                mgmt_port_id)

    def test_delete_resource_port_fail_only_twice(self):
        mgmt_port_id = 'fake_port_id'
        mocked_plugin = mock.MagicMock()
        mock_ctx = mock.MagicMock()
        mocked_plugin.delete_port = mock.MagicMock(
            side_effect=[n_exc.NeutronException, n_exc.NeutronException,
                         mock.Mock])
        with mock.patch.object(VIFHotPlugPluggingDriver,
                               '_core_plugin') as plugin:
            plugin.__get__ = mock.MagicMock(return_value=mocked_plugin)
            vif_plugging_driver = VIFHotPlugPluggingDriver()
            vif_plugging_driver._delete_resource_port(mock_ctx, mgmt_port_id)
            self.assertEqual(3, mocked_plugin.delete_port.call_count)

    def test_delete_resource_port_handle_port_not_found(self):
        mgmt_port_id = 'fake_port_id'
        mocked_plugin = mock.MagicMock()
        mock_ctx = mock.MagicMock()
        mocked_plugin.delete_port = mock.MagicMock(
            side_effect=n_exc.PortNotFound(port_id=mgmt_port_id))
        with mock.patch.object(VIFHotPlugPluggingDriver,
                               '_core_plugin') as plugin:
            plugin.__get__ = mock.MagicMock(return_value=mocked_plugin)
            vif_plugging_driver = VIFHotPlugPluggingDriver()
            vif_plugging_driver._delete_resource_port(mock_ctx, mgmt_port_id)
            self.assertEqual(1, mocked_plugin.delete_port.call_count)

    @mock.patch.object(VIFHotPlugPluggingDriver, '_dev_mgr')
    def test_setup_logical_port_connectivity(self, mock_svc_vm_mgr):
        hosting_port_obj = mock.MagicMock(id='hosting_port_id')
        hosting_info_obj = mock.MagicMock(hosting_port=hosting_port_obj)
        mock_portdb = mock.MagicMock(hosting_info=hosting_info_obj)

        hosting_device_id = 'fake_hosting_device_id'
        mocked_plugin = mock.MagicMock()
        mock_ctx = mock.MagicMock()
        with mock.patch.object(VIFHotPlugPluggingDriver,
                               '_core_plugin') as plugin:
            plugin.__get__ = mock.MagicMock(return_value=mocked_plugin)
            vif_plugging_driver = VIFHotPlugPluggingDriver()
            vif_plugging_driver.setup_logical_port_connectivity(
                mock_ctx, mock_portdb, hosting_device_id)
            vif_plugging_driver._dev_mgr.svc_vm_mgr.interface_attach\
                .assert_called_once_with(hosting_device_id,
                mock_portdb.hosting_info.hosting_port.id)

    def test_create_hosting_device_resources(self):
        complementary_id = 'fake_complementary_id'
        tenant_id = 'fake_tenantid'
        mgmt_context = {'mgmt_nw_id': 'fake_mgmt_nw_id',
                        'mgmt_sec_grp_id': 'fake_mgmt_sec_grp_id'}
        max_hosted = 'fake_max_hosted'
        mocked_plugin = mock.MagicMock()
        mock_ctx = mock.MagicMock()
        with mock.patch.object(VIFHotPlugPluggingDriver,
                               '_core_plugin') as plugin:
            plugin.__get__ = mock.MagicMock(return_value=mocked_plugin)
            vif_plugging_driver = VIFHotPlugPluggingDriver()
            vif_plugging_driver.create_hosting_device_resources(
                mock_ctx, complementary_id, tenant_id, mgmt_context,
                max_hosted)
            self.assertEqual(True, mocked_plugin.create_port.called)
            self.assertEqual(1, mocked_plugin.create_port.call_count)

    def test_create_hosting_device_resources_exception(self):
        complementary_id = 'fake_complementary_id'
        tenant_id = 'fake_tenantid'
        mgmt_context = {'mgmt_nw_id': 'fake_mgmt_nw_id',
                        'mgmt_sec_grp_id': 'fake_mgmt_sec_grp_id'}
        max_hosted = 'fake_max_hosted'
        mock_delete_resources = mock.MagicMock()
        mocked_plugin = mock.MagicMock()
        mocked_plugin.create_port = mock.MagicMock(
            side_effect=n_exc.NeutronException)

        mock_ctx = mock.MagicMock()
        with mock.patch.object(VIFHotPlugPluggingDriver,
                               '_core_plugin') as plugin:
            plugin.__get__ = mock.MagicMock(return_value=mocked_plugin)
            vif_plugging_driver = VIFHotPlugPluggingDriver()
            vif_plugging_driver.delete_hosting_device_resources = (
                mock_delete_resources)
            result = vif_plugging_driver.create_hosting_device_resources(
                mock_ctx, complementary_id, tenant_id, mgmt_context,
                max_hosted)
            self.assertEqual(True, mocked_plugin.create_port.called)
            self.assertEqual(1, mocked_plugin.create_port.call_count)
            self.assertEqual(True, mock_delete_resources.called)
            self.assertEqual(1, mock_delete_resources.call_count)
            self.assertIsNone(result['mgmt_port'])
            self.assertEqual([], result['ports'])



# Copyright 2016 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


import eventlet
import os
import signal
import socket
import struct
import time
import uuid

from neutron.common import config
from neutron_lib import exceptions
from oslo_config import cfg
from oslo_log import log as logging

from networking_cisco._i18n import _
from networking_cisco import backwards_compatibility as bc
from networking_cisco.plugins.cisco.cpnr.cpnr_client import UnexpectedError
from networking_cisco.plugins.cisco.cpnr import debug_stats
from networking_cisco.plugins.cisco.cpnr import netns

LOG = logging.getLogger(__name__)

DEBUG_STATS_MIN_WRITE_INTERVAL = 30
MONITOR_INTERVAL = 1
CLEANUP_INTERVAL = 30
STALE_REQUEST_TMO = 120
RECV_BUFFER_SIZE = 4096
NS_RELAY_PENDING = 'NS_RELAY_PENDING'
NS_RELAY_RUNNING = 'NS_RELAY_RUNNING'
NS_RELAY_DELETING = 'NS_RELAY_DELETING'
NS_PREFIX = 'qdhcp'
DNS_SERVER_PORT = 53
RLIMIT_NOFILE_LIMIT = 16384

opt_cpnr_group = cfg.OptGroup(
    name='cisco_pnr', title='Cisco Prime Network Registrar configuration')

OPTS = [
    cfg.StrOpt('external_interface',
               default='lo',
               help=_('Interface for communicating with DHCP/DNS server.')),
    cfg.StrOpt('dns_server_addr',
               default='127.0.0.1',
               help=_('DNS server IP address.')),
    cfg.IntOpt('dns_server_port',
               default=53,
               help=_('DNS server UDP port number.')),
    cfg.BoolOpt('enable_dns_stats',
                default=False,
                help=_('Enable DNS stats.')),
    cfg.IntOpt('dns_stats_interval',
               default=60,
               help=_('DNS stats polling interval.'))
]


class DnsRelayAgent(object):
    """Relay DNS packets between neutron networks and external DNS server.

    Receives unicast DNS requests via sockets which are opened in each neutron
    network namespace.  Additional DNS options are appended to the request
    to indicate from which network the request originated. Requests are then
    forwarded to the configured DNS server address.

    Receives unicast DNS responses from the DNS server via socket opened in
    the global network namespace.  Additional options are stripped from the
    response. The response is then forwarded to the originating network.
    """

    def __init__(self):
        self.conf = cfg.CONF
        self.conf.register_group(opt_cpnr_group)
        self.conf.register_opts(OPTS, opt_cpnr_group)
        if bc.NEUTRON_VERSION < bc.NEUTRON_NEWTON_VERSION:
            LOG.debug('(cisco_pnr) http_server: {}'.format(
                self.conf.cisco_pnr.http_server))
        LOG.debug('(cisco_pnr) external_interface: {}'.format(
            self.conf.cisco_pnr.external_interface))
        LOG.debug('(cisco_pnr) dhcp_server_addr: {}'.format(
            self.conf.cisco_pnr.dns_server_addr))
        LOG.debug('(cisco_pnr) dhcp_server_port: {}'.format(
            self.conf.cisco_pnr.dns_server_port))
        self.ns_states = {}
        self.request_info_by_msgid = {}
        self.ext_sock = None
        self.ext_addr = ""
        self.ns_lock = eventlet.semaphore.Semaphore()
        self.debug_stats = debug_stats.DebugStats('dns')
        self.kill_now = False

    def serve(self):
        self.greenpool = eventlet.GreenPool(4)
        self.greenpool.spawn_n(self._server_network_relay)
        self.greenpool.spawn_n(self._namespace_monitor)
        self.greenpool.spawn_n(self._cleanup_stale_requests)
        if self.conf.cisco_pnr.enable_dns_stats:
            self.greenpool.spawn_n(self._write_debug_stats)
        self.greenpool.waitall()

    def _signal_handler(self, signum, frame):
        LOG.debug('Recieved the signal %s', signum)
        self.kill_now = True

    def _namespace_monitor(self):

        while True:
            if self.kill_now:
                break
            eventlet.sleep(MONITOR_INTERVAL)

            # Get list of network namespaces on system
            try:
                curr_ns = set(netns.nslist())
            except Exception:
                LOG.error('Failed to get current namespace set')
                continue

            # For each unknown namespace, start a relay thread
            for ns in curr_ns:
                if not ns.startswith(NS_PREFIX) or ns in self.ns_states:
                    continue
                self.ns_states[ns] = NS_RELAY_PENDING
                eventlet.spawn_n(self._client_network_relay, ns)

            # Set state to DELETING for any unknown namespaces
            for ns in self.ns_states:
                if ns in curr_ns:
                    continue
                self.ns_states[ns] = NS_RELAY_DELETING
        LOG.debug('Namespace Monitor exiting')

    def _server_network_relay(self):

        # Open a socket in the global namespace for DNS
        try:
            self.ext_sock, self.ext_addr, ext_port = (
                self._open_dns_ext_socket())
        except Exception:
            LOG.exception('Failed to open dns external '
                          'socket in global ns')
            return
        recvbuf = bytearray(RECV_BUFFER_SIZE)
        LOG.debug("Opened dns external server socket on addr:%s:%i",
                  self.ext_addr, ext_port)

        # Forward DNS responses from external to internal networks
        while True:
            if self.kill_now:
                break
            try:
                self.ext_sock.settimeout(1)
                size = self.ext_sock.recv_into(recvbuf)
                pkt = DnsPacket.parse(recvbuf, size)
                msgid = pkt.get_msgid()
                LOG.debug("got dns response pkt, msgid =  %i", msgid)
                if msgid not in self.request_info_by_msgid:
                    LOG.debug('Could not find request by msgid %i', msgid)
                    continue
                int_sock, int_addr, int_port, createtime, viewid = (
                    self.request_info_by_msgid[msgid])
                self.debug_stats.increment_pkts_from_server(viewid)
                LOG.debug("forwarding response to internal namespace "
                          "at %s:%i", int_addr, int_port)
                int_sock.sendto(recvbuf[:size], (int_addr, int_port))
                del self.request_info_by_msgid[msgid]
                self.debug_stats.increment_pkts_to_client(viewid)
            except socket.timeout:
                pass
            except Exception:
                LOG.exception('Failed to forward dns response')
        LOG.debug('Server Network relay exiting')

    def _client_network_relay(self, namespace):

        # Open a socket in the DNS network namespace
        try:
            with self.ns_lock, netns.Namespace(namespace):
                int_sock, int_addr, int_port = self._open_dns_int_socket()
        except exceptions.BaseException:
            LOG.exception('Failed to open dns server socket in %s',
                          namespace)
            del self.ns_states[namespace]
            return
        self.ns_states[namespace] = NS_RELAY_RUNNING
        recvbuf = bytearray(RECV_BUFFER_SIZE)
        LOG.debug("Opened dns server socket on ns: %s, addr:%s:%i",
                  namespace, int_addr, int_port)

        # Convert the namespace into a view id
        viewid = self._convert_namespace_to_viewid(namespace)

        self.debug_stats.add_network_stats(viewid)

        # Forward DNS requests from internal to external networks
        while self.ns_states[namespace] != NS_RELAY_DELETING:
            if self.kill_now:
                break
            try:
                int_sock.settimeout(1)
                size, (src_addr, src_port) = int_sock.recvfrom_into(recvbuf)
                LOG.debug("got dns request from ns: %s", namespace)
                self.debug_stats.increment_pkts_from_client(viewid)
                pkt = DnsPacket.parse(recvbuf, size)
                pkt.set_viewid(viewid)

                # Store off some state to know where to forward response later
                msgid = pkt.get_msgid()
                createtime = time.time()
                self.request_info_by_msgid[msgid] = [int_sock,
                                    src_addr, src_port, createtime, viewid]
                LOG.debug("forwarding request to external nameserver")
                self.ext_sock.send(pkt.data())
                self.debug_stats.increment_pkts_to_server(viewid)
            except socket.timeout:
                pass
            except Exception:
                LOG.exception('Failed to forward dns request to server '
                              'from %s', namespace)

        # Cleanup socket and internal state
        try:
            del self.ns_states[namespace]
            self.debug_stats.del_network_stats(viewid)
            int_sock.close()
        except Exception:
            LOG.warning('Failed to cleanup dns relay for %s', namespace)
        LOG.debug('Client network relay exiting')

    def _open_dns_ext_socket(self):

        # find configured external interface ip address
        for ifname, addr, mask in netns.iflist():
            if ifname == self.conf.cisco_pnr.external_interface:
                break
        else:
            raise UnexpectedError(msg='Failed to find external '
                                      'interface matching config')

        # open, bind, and connect UDP socket
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.bind((addr, 0))    # any available port
        addr, port = s.getsockname()
        s.connect((self.conf.cisco_pnr.dns_server_addr,
                   self.conf.cisco_pnr.dns_server_port))
        return s, addr, port

    def _open_dns_int_socket(self):

        # list interfaces, make sure there is at least one
        interfaces = netns.iflist(ignore=("lo",))
        if not interfaces:
            raise UnexpectedError(msg="Failed to find "
                                  "single interface in dhcp namespace")
        ifname, addr, mask = interfaces[0]

        # open socket for receiving DNS requests and sending DNS responses
        # on internal net
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        s.bind((addr, DNS_SERVER_PORT))

        return s, addr, DNS_SERVER_PORT

    def _cleanup_stale_requests(self):

        while True:
            if self.kill_now:
                break
            eventlet.sleep(CLEANUP_INTERVAL)

            currtime = time.time()
            for msgid in self.request_info_by_msgid.keys():
                reqvals = self.request_info_by_msgid[msgid]
                createtime = reqvals[3]
                if (currtime - createtime) > STALE_REQUEST_TMO:
                    del self.request_info_by_msgid[msgid]
        LOG.debug('Cleanup thread exiting')

    def _convert_namespace_to_viewid(self, namespace):
        netid = namespace[6:]
        netid = uuid.UUID(netid)
        return str(netid.int & 0x7fffffff)

    def _write_debug_stats(self):
        polling_interval = max(DEBUG_STATS_MIN_WRITE_INTERVAL,
                               self.conf.cisco_pnr.dns_stats_interval)
        while True:
            eventlet.sleep(polling_interval)
            self.debug_stats.write_stats_to_file()


class DnsPacket(object):

    # Byte array representation of common TXT RR fields that will be inserted
    # in each request
    #    Format is:
    #       - domain name ("10_cpnr_info5cisco3com0")
    #       - rr type (2 bytes, value = 16)
    #       - class (2 bytes, value = 1)
    #       - ttl (4 bytes, value = 0)
    TXT_RR = bytearray(b'\x0a\x5f\x63\x70\x6e\x72\x5f\x69\x6e\x66\x6f\x05'
                       b'\x63\x69\x73\x63\x6f\x03\x63\x6f\x6d\x00\x00\x10'
                       b'\x00\x01\x00\x00\x00\x00')

    QUERY_TYPE_AND_CLASS = 4
    TYPE_CLASS_AND_TTL_LENGTH = 8
    OPTIONAL_RR = 41
    COUNTS_LENGTH = 8  # Question, Answer, Authority and Additional RR Count.
    IDENTIFIER_FLAGS_AND_CODES_LENGTH = 4

    def __init__(self):
        self.buf = ''
        self.msgid = ''
        self.isreq = False
        self.viewid = ""
        self.arcnt = 0
        self.txt_insert_pos = -1
        self.optlen = 0

    @classmethod
    def parse(cls, buf, buflen):
        pkt = DnsPacket()

        # parse out the header
        (pkt.msgid,) = cls.struct('!H').unpack_from(buf, 0)
        (info,) = cls.struct('B').unpack_from(buf, 2)
        pos = DnsPacket.IDENTIFIER_FLAGS_AND_CODES_LENGTH

        # check if query
        isquery = not (info & 0x80)
        if not isquery:
            pkt.buf = buf
            LOG.debug("DNS packet is a response")
            return pkt

        LOG.debug("DNS packet is a query")
        pkt.isreq = True

        (qdcnt,) = cls.struct('!H').unpack_from(buf, 4)
        (ancnt,) = cls.struct('!H').unpack_from(buf, 6)
        (nscnt,) = cls.struct('!H').unpack_from(buf, 8)
        (arcnt,) = cls.struct('!H').unpack_from(buf, 10)
        pkt.arcnt = arcnt
        pos += DnsPacket.COUNTS_LENGTH

        LOG.debug('Parsed pkt: msgid %s qdcnt %i ancnt %i nscnt %i '
                  'arcnt %i', pkt.msgid, qdcnt, ancnt, nscnt, arcnt)

        for i in range(qdcnt):
            pos = cls.skip_over_domain_name(buf, pos)
            pos += DnsPacket.QUERY_TYPE_AND_CLASS

        if ancnt != 0 or nscnt != 0:
            # unexpected, log and return packet
            LOG.debug('Unexpected answers in query, ancnt %i nscnt %i',
                      ancnt, nscnt)
            pkt.buf = buf
            return pkt

        # walk through additional section, check for OPT RR (if present, must
        # come last)
        tmp_pos = pos
        for i in range(arcnt):
            tmp_pos = cls.skip_over_domain_name(buf, tmp_pos)
            (type,) = cls.struct('!H').unpack_from(buf, tmp_pos)
            tmp_pos += DnsPacket.TYPE_CLASS_AND_TTL_LENGTH
            (rdlen,) = cls.struct('!H').unpack_from(buf, tmp_pos)
            tmp_pos += 2 + rdlen  # rdlength and rdata

            if type == DnsPacket.OPTIONAL_RR:
                pkt.optlen = buflen - pos
                break
            else:
                pos = tmp_pos
        pkt.txt_insert_pos = pos
        pkt.buf = buf
        return pkt

    def get_msgid(self):
        return self.msgid

    def set_viewid(self, id):
        self.viewid = id

    def data(self):
        if not self.isreq or not self.viewid:
            return self.buf

        # make a copy of OPT RR, if present
        opt_data = ''
        if self.optlen != 0:
            opt_data = self.buf[self.txt_insert_pos:self.txt_insert_pos +
                                self.optlen]

        # insert TXT RR and data into buf
        pos = self.txt_insert_pos
        self.buf[pos:pos + len(DnsPacket.TXT_RR)] = DnsPacket.TXT_RR
        pos += len(DnsPacket.TXT_RR)
        txt_str = 'view: %s' % (self.viewid,)
        self.struct('!HB%is' %
                    (len(txt_str),)).pack_into(self.buf, pos,
                                               len(txt_str) + 1,
                                               len(txt_str),
                                               txt_str.encode('utf-8'))

        pos += 3 + len(txt_str)

        # bump up arcnt
        self.arcnt += 1
        self.struct('!H').pack_into(self.buf, 10, self.arcnt)

        # copy OPT RR back in at end if presesnt
        if opt_data:
            self.buf[pos:pos + len(opt_data)] = opt_data
            pos += len(opt_data)

        return self.buf[:pos]

    @classmethod
    def skip_over_domain_name(cls, buf, pos):
        tmplen = -1
        while tmplen != 0:
            (tmplen,) = cls.struct('B').unpack_from(buf, pos)
            if (tmplen & 0x80) and (tmplen & 0x40):
                pos += 2  # length and pointer, comes last
                break
            else:
                pos += 1 + tmplen
        return pos

    structcache = {}

    @classmethod
    def struct(cls, fmt):
        return cls.structcache.setdefault(fmt, struct.Struct(fmt))


def main():
    try:
        netns.increase_ulimit(RLIMIT_NOFILE_LIMIT)
    except Exception:
        LOG.error('Failed to increase ulimit for DNS relay')
    if os.getuid() != 0:
        config.setup_logging()
        LOG.error('Must run dns relay as root')
        return
    eventlet.monkey_patch()
    config.setup_logging()
    relay = DnsRelayAgent()
    signal.signal(signal.SIGINT, relay._signal_handler)
    signal.signal(signal.SIGTERM, relay._signal_handler)
    relay.serve()

if __name__ == "__main__":
    main()



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Exceptions by Cisco Configuration Agent."""

from neutron_lib import exceptions

from networking_cisco._i18n import _


class DriverException(exceptions.NeutronException):
    """Exception created by the Driver class."""


class DriverExpectedKeyNotSetException(DriverException):
    """An attribute expected to be set by plugin is missing"""
    message = (_("Value for expected key: %(key)s is missing."
                 "Driver cannot proceed"))


class InitializationException(DriverException):
    """Exception when initialization of Routing Driver object."""
    message = (_("Critical device parameter missing. Failed initializing "
                 "routing driver object."))


class ConnectionException(DriverException):
    """Connection exception when connecting to IOS XE hosting device."""
    message = (_("Failed connecting to Device. Reason: %(reason)s. "
                 "Connection params are User:%(user)s, Host:%(host)s, "
                 "Port:%(port)s, Device timeout:%(timeout)s."))


class IOSXEConfigException(DriverException):
    """Configuration exception thrown when modifying the running config."""
    message = (_("Error executing snippet:%(snippet)s. "
                 "Hosting device:%(dev_id)s Mgmt IP:%(ip)s "
                 "ErrorType:%(type)s ErrorTag:%(tag)s Config string:%("
                 "confstr)s."))


class IOSXEMissingInterfaceException(DriverException):
    """Configuration exception thrown when modifying the running config."""
    message = (_("Interface corresponding to port:%(id)s and mac-address:%("
                 "mac)s is missing in the device. Cannot proceed with "
                 "interface config."))


class IOSXEUnknownValueException(DriverException):
    """IOS XE device exception thrown when an unknown value is received."""
    message = (_("Data in attribute: %(attribute)s does not correspond to "
                 "expected value. Value received is %(value)s. "))


class DriverNotExist(DriverException):
    message = _("Driver %(driver)s does not exist.")


class DriverNotFound(DriverException):
    message = _("Driver not found for %(resource)s id:%(id)s.")


class DriverNotSetForMissingParameter(DriverException):
    message = _("Driver cannot be set for missing parameter:%(p)s.")


class HAParamsMissingException(DriverException):
    """MissingParams exception thrown when HA params are missing"""
    message = (_("For router: %(r_id)s and port: %(p_id)s, HA_ENABLED is set, "
                 "but port ha info is missing. Port details: %(port)s"))



# Copyright 2015 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


import sys

from oslo_config import cfg

from networking_cisco._i18n import _LE


from networking_cisco.apps.saf.agent.vdp import (
    lldpad_constants as vdp_const)
from networking_cisco.apps.saf.common import constants as com_const
from networking_cisco.apps.saf.common import utils
from networking_cisco.apps.saf.server.services import constants as const
from networking_cisco.apps.saf.server.services.firewall.native import (
    fw_constants as fw_const)


default_keystone_opts = {
    'keystone_authtoken': {
        'username': 'admin',
        'project_name': 'admin',
        'user_domain_name': 'default',
        'project_domain_name': 'default',
    },
}

default_neutron_opts = {
    'neutron': {
        'username': 'neutron',
        'project_name': 'service',
        'user_domain_name': 'default',
        'project_domain_name': 'default',
    },
}

default_nova_opts = {
    'nova': {
        'username': 'nova',
        'project_name': 'service',
        'user_domain_name': 'default',
        'project_domain_name': 'default',
        'region_name': 'RegionOne',
        'api_version': '2.1',
    },
}


default_dfa_agent_opts = {
    'dfa_agent': {
        'integration_bridge': 'br-int',
        'external_dfa_bridge': 'br-ethd',
    },
}

default_vdp_opts = {
    'vdp': {
        'mgrid2': vdp_const.VDP_MGRID,
        'typeid': vdp_const.VDP_TYPEID,
        'typeidver': vdp_const.VDP_TYPEID_VER,
        'vsiidfrmt': vdp_const.VDP_VSIFRMT_UUID,
        'hints': 'none',
        'filter': vdp_const.VDP_FILTER_GIDMACVID,
        'vdp_sync_timeout': vdp_const.VDP_SYNC_TIMEOUT,
    },
}

default_firewall_opts = {
    'firewall': {
        'device': fw_const.DEVICE,
        'sched_policy': fw_const.SCHED_POLICY,
        'fw_auto_serv_nwk_create': fw_const.AUTO_NWK_CREATE,
        'fw_service_host_profile': fw_const.HOST_PROF,
        'fw_service_host_fwd_mode': fw_const.HOST_FWD_MODE,
        'fw_service_part_vrf_profile': fw_const.PART_PROF,
        'fw_service_ext_profile': fw_const.EXT_PROF,
        'fw_service_ext_fwd_mode': fw_const.EXT_FWD_MODE,
        'fw_service_in_ip_start': fw_const.IN_IP_START,
        'fw_service_in_ip_end': fw_const.IN_IP_END,
        'fw_service_out_ip_start': fw_const.OUT_IP_START,
        'fw_service_out_ip_end': fw_const.OUT_IP_END,
        'fw_service_dummy_ip_subnet': fw_const.DUMMY_IP_SUBNET,
    },
}

DEFAULT_LOG_LEVELS = (
    "amqp=WARN, amqplib=WARN, oslo.messaging=WARN, pika=WARN, paramiko=WARN,"
    "paramiko.transport=WARN,"
    "paramiko.transport.sftp=WARN,"
    "pika.callback=WARN,oslo.messaging._drivers=WARN"
)

default_log_opts = {
    'dfa_log': {
        'use_syslog': 'False',
        'syslog_lgo_facility': 'LOG_USER',
        'log_dir': '.',
        'log_file': 'fabric_enabler.log',
        'log_level': 'WARNING',
        'log_format': '%(asctime)s %(levelname)8s [%(name)s] %(message)s',
        'log_date_format': '%Y-%m-%d %H:%M:%S',
        'default_log_levels': DEFAULT_LOG_LEVELS,
    },
}

default_sys_opts = {
    'sys': {
        'root_helper': 'sudo',
    },
}

default_dcnm_opts = {
    'dcnm': {
        'default_cfg_profile': 'defaultNetworkIpv4EfProfile',
        'default_vrf_profile': 'vrf-common-universal',
        'default_partition_name': 'CTX',
        'dcnm_net_ext': '(DCNM)',
        'gateway_mac': '20:20:00:00:00:AA',
        'dcnm_dhcp_leases': '/var/lib/dhcpd/dhcpd.leases',
        'dcnm_dhcp': 'false',
        'segmentation_reuse_timeout': com_const.SEG_REUSE_TIMEOUT,
        'vlan_id_min': const.VLAN_ID_MIN,
        'vlan_id_max': const.VLAN_ID_MAX,
        'vlan_reuse_timeout': const.VLAN_REUSE_TIMEOUT,
        'orchestrator_id': com_const.ORCHESTRATOR_ID
    },
}

default_notify_opts = {
    'dfa_notify': {
        'cisco_dfa_notify_queue': 'cisco_dfa_%(service_name)s_notify',
    },
}

default_loadbalance_opts = {
    'loadbalance': {
        'lb_enabled': 'False',
        'lb_native': 'True',
    },
}

default_opts_list = [
    default_log_opts,
    default_neutron_opts,
    default_nova_opts,
    default_keystone_opts,
    default_dfa_agent_opts,
    default_vdp_opts,
    default_sys_opts,
    default_dcnm_opts,
    default_notify_opts,
    default_firewall_opts,
    default_loadbalance_opts,
]


class CiscoDFAConfig(object):

    """Cisco DFA Mechanism Driver Configuration class."""

    def __init__(self, service_name=None):
        self.dfa_cfg = {}
        self._load_default_opts()
        args = sys.argv[1:]
        try:
            opts = [(args[i], args[i + 1]) for i in range(0, len(args), 2)]
        except IndexError:
            opts = []

        cfgfile = cfg.find_config_files(service_name)
        for k, v in opts:
            if k == '--config-file':
                cfgfile.insert(0, v)
        multi_parser = cfg.MultiConfigParser()
        read_ok = multi_parser.read(cfgfile)

        if len(read_ok) != len(cfgfile):
            raise cfg.Error(_LE("Failed to parse config file."))

        for parsed_file in multi_parser.parsed:
            for parsed_item in parsed_file.keys():
                if parsed_item not in self.dfa_cfg:
                    self.dfa_cfg[parsed_item] = {}
                for key, value in parsed_file[parsed_item].items():
                    self.dfa_cfg[parsed_item][key] = value[0]

        # Convert it to object.
        self._cfg = utils.Dict2Obj(self.dfa_cfg)

    def _load_default_opts(self):
        """Load default options."""

        for opt in default_opts_list:
            self.dfa_cfg.update(opt)

    @property
    def cfg(self):
        return self._cfg



# Copyright 2015 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""CLI module for fabric enabler."""

from __future__ import print_function

import cmd
import itertools
import pkg_resources
import platform
from prettytable import PrettyTable
import six
import sys

from oslo_serialization import jsonutils
from oslo_utils import netutils

from networking_cisco.apps.saf.common import config
from networking_cisco.apps.saf.common import constants
from networking_cisco.apps.saf.common import dfa_exceptions as dexc
from networking_cisco.apps.saf.common import rpc
from networking_cisco.apps.saf.common import utils
from networking_cisco.apps.saf.server import cisco_dfa_rest as cdr


class DfaCli(cmd.Cmd):

    """Represents fabric enabler command line interface."""

    prompt = '(enabler) '
    intro = 'Fabric Enabler Command Line Interface'

    def __init__(self):
        self.ctl_host = platform.node()
        cmd.Cmd.__init__(self)
        self._cfg = config.CiscoDFAConfig().cfg
        self.dcnm_client = cdr.DFARESTClient(self._cfg)
        self.setup_client_rpc()
        self.clnt = None

    def setup_client_rpc(self):
        url = self._cfg.dfa_rpc.transport_url % (
            {'ip': self.ctl_host})
        self.clnt = rpc.DfaRpcClient(url, constants.DFA_SERVER_QUEUE)

    def set_static_ip_address(self, ipaddr, macaddr):
        context = {}
        args = jsonutils.dumps(dict(mac=macaddr, ip=ipaddr))
        msg = self.clnt.make_msg('set_static_ip_address', context, msg=args)
        resp = self.clnt.cast(msg)
        return resp

    def do_set_static_ip(self, line):
        args = line.split()
        ip_mac = dict(itertools.izip_longest(args[::2], args[1::2],
                                             fillvalue=''))
        ipaddr = ip_mac.get('--ip')
        macaddr = ip_mac.get('--mac')
        # Some sanity check.
        if (not ipaddr or not macaddr or
                not utils.is_valid_ipv4(ipaddr)
                or not netutils.is_valid_mac(macaddr)):
            print('Invalid input parameters.\n'
                  'Usage:'
                  ' set_static_ip --mac <mac address> --ip <ip address>')
            return

        self.set_static_ip_address(ipaddr, macaddr)

    def do_get_config_profile(self, line):
        try:
            cfgp_list = self.dcnm_client.config_profile_list()
            if not cfgp_list:
                print('No config profile found.')
                return
        except dexc.DfaClientRequestFailed:
            print('Failed to access DCNM.')
            return

        cfg_table = PrettyTable(['Config Profile Name', 'Alias'])
        for cfg in cfgp_list:
            if cfg.startswith('defaultNetwork'):
                cfg_alias = cfg.split('defaultNetwork')[1].split('Profile')[0]
            elif cfg.endswith('Profile'):
                cfg_alias = cfg.split('Profile')[0]
            else:
                cfg_alias = cfg
            cfg_table.add_row([cfg, cfg_alias])

        print(cfg_table)

    def do_list_networks(self, line):
        tenant_name = line
        if not tenant_name:
            print('Tenant name is required.')
            return

        try:
            part_name = self._cfg.dcnm.default_partition_name
            net_list = self.dcnm_client.list_networks(tenant_name, part_name)
            if not net_list:
                print('No network found.')
                return
        except dexc.DfaClientRequestFailed:
            print('Failed to access DCNM.')
            return

        list_table = None
        for net in net_list:
            columns = net.keys()
            if list_table is None:
                list_table = PrettyTable(columns)
            if list_table:
                list_table.add_row(net.values())

        print(list_table)

    def do_get_network(self, line):
        args = line.split()
        if len(args) < 2:
            print('Invalid parameters')
            return

        if not args[1].isdigit():
            print('Invalid segmentation id %s.', args[1])
            return

        try:
            net = self.dcnm_client.get_network(args[0], args[1])
            if not net:
                print('No network found.')
                return
        except dexc.DfaClientRequestFailed:
            print('Failed to access DCNM.')
            return

        net_table = PrettyTable(net.keys())
        row = []
        for key, val in six.iteritems(net):
            if key == 'configArg' or key == 'dhcpScope':
                val = str(val)
            row.append(val)
        net_table.add_row(row)

        print(net_table)

    def do_list_organizations(self, line):
        '''Get list of organization on DCNM.'''

        org_list = self.dcnm_client.list_organizations()
        if not org_list:
            print('No organization found.')
            return

        org_table = PrettyTable(['Organization Name'])
        for org in org_list:
            org_table.add_row([org['organizationName']])

        print(org_table)

    def do_get_dcnm_version(self, line):
        '''Get current version of DCNM.'''

        ver = self.dcnm_client.get_version()

        print(ver)

    def do_get_enabler_version(self, line):
        '''Get current fabric enabler's package version.'''

        print('Version: %s' % pkg_resources.get_distribution(
            "networking-cisco").version)

    def help_get_config_profile(self):
        print('\n'.join(['get_config_profile',
                         'Display supported configuration profile in DCNM']))

    def help_list_networks(self):
        print('\n'.join(['list_networks tenant-name',
                         'Display list of network for given tenant.']))

    def help_get_network(self):
        print('\n'.join(['get_network tenant-name segmentation_id',
                         'Display network details.']))

    def help_set_static_ip(self):
        print('\n'.join(['set_static_ip --mac <mac address> --ip <ip address>',
                         'Set static ip address for an instance.']))

    def emptyline(self):
        return

    def do_prompt(self, line):
        '''Set prompt for the command line.'''

        self.prompt = line + ' '

    def do_quit(self, line):
        '''exit the program.'''

        sys.exit(1)

    def do_EOF(self, line):
        '''Use Ctrl-D to exit the program.'''

        return True

    # Shortcuts
    do_q = do_quit


def dfa_cli():
    # Add default config file.
    if len(sys.argv[1:]) % 2:
        sys.argv.append("")
    sys.argv.append('--config-file')
    sys.argv.append('/etc/saf/enabler_conf.ini')
    DfaCli().cmdloop()

if __name__ == '__main__':
    sys.exit(dfa_cli())



# Copyright 2016 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

import mock
from oslo_log import log as logging
from oslo_utils import fileutils
from oslo_utils import uuidutils
import webob.exc

from neutron.common import test_lib
from neutron.tests.unit.extensions import test_l3

from networking_cisco import backwards_compatibility as bc
from networking_cisco.plugins.cisco.common import cisco_constants
from networking_cisco.plugins.cisco.device_manager.plugging_drivers import (
    aci_vlan_trunking_driver as aci_vlan)
from networking_cisco.plugins.cisco.extensions import routerrole
from networking_cisco.tests.unit.cisco.l3 import (
    test_l3_router_appliance_plugin)

_uuid = uuidutils.generate_uuid
LOG = logging.getLogger(__name__)

ROUTER_ROLE_ATTR = routerrole.ROUTER_ROLE_ATTR

DEVICE = 'mydev'
PORT_ID = 'myportid'
MAC_ADDRESS = '00:11:22:33:44:55'
APP_PROFILE = 'myAppProfile'
SEGMENT_ID = '11'
NETWORK_TYPE = 'opflex'
TENANT = 'mytenent'
HOST = 'ubuntu'
NETWORK_TENANT = 'net_tenant'
EPG_NAME = 'myEpg'
APIC_VLAN1 = 11
APIC_VLAN2 = 12


class FakePortDb(object):

    def __init__(self, id, network_id, device_owner, device_id):
        self.id = id
        self.network_id = network_id
        self.device_id = device_id
        self.device_owner = device_owner
        self.hosting_info = {}
        self.extra_subnets = []

    def get(self, name):
        return self[name]

    def __getitem__(self, key):
        if key == 'id':
            return self.id
        if key == 'network_id':
            return self.network_id
        if key == 'device_owner':
            return self.device_owner
        if key == 'device_id':
            return self.device_id
        if key == 'extra_subnets':
            return self.extra_subnets
        if key == 'hosting_info':
            return self.hosting_info


class TestAciVLANTrunkingPlugDriverBase(
    test_l3_router_appliance_plugin.L3RouterApplianceTestCaseBase,
    test_l3.L3NatTestCaseMixin):
    """Test class for Base ACI VLAN Trunking Plugging driver

    This class tests the functionality of the ACI VLAN Trunking Plugging
    driver, which is indpendent of the workflow used (GBP or Neutron)
    """

    # we use router types defined in .ini file.
    configure_routertypes = False
    router_type = 'ASR1k_Neutron_router'

    def setUp(self):
        super(TestAciVLANTrunkingPlugDriverBase, self).setUp(
            create_mgmt_nw=False)
        # save possible test_lib.test_config 'config_files' dict entry so we
        # can restore it after tests since we will change its value
        self._old_config_files = copy.copy(test_lib.test_config.get(
            'config_files'))
        # include config files for device manager service plugin and router
        # service plugin since we define a number of hosting device templates,
        # hosting devices and routertypes there
        self._add_device_manager_plugin_ini_file()
        self._add_router_plugin_ini_file()
        #TODO(bobmel): Fix bug in test_extensions.py and we can remove the
        # below call to setup_config()
        self.setup_config()
        self.l3_plugin._core_plugin.mechanism_manager = mock.MagicMock()
        plug = aci_vlan.AciVLANTrunkingPlugDriver()
        plug._apic_driver = mock.Mock()
        self.plugging_driver = plug
        self.vlan_dict = {'net1': APIC_VLAN1, 'net2': APIC_VLAN2}

    def tearDown(self):
        if self._old_config_files is None:
            test_lib.test_config.pop('config_files', None)
        else:
            test_lib.test_config['config_files'] = self._old_config_files
        super(TestAciVLANTrunkingPlugDriverBase, self).tearDown()

    def test_create_hosting_device_resources(self):
        tenant_id = 'some_tenant_id'
        ctx = bc.context.Context('', tenant_id, is_admin=True)
        mgmt_context = {'mgmt_nw_id': None}
        res = self.plugging_driver.create_hosting_device_resources(
            ctx, "some_id", tenant_id, mgmt_context, 2)
        self.assertIsNone(res['mgmt_port'])
        self.assertEqual(1, len(res))

    def test_create_hosting_device_resources_no_mgmt_context(self):
        tenant_id = 'some_tenant_id'
        ctx = bc.context.Context('', tenant_id, is_admin=True)
        res = self.plugging_driver.create_hosting_device_resources(
            ctx, "some_id", tenant_id, None, 2)
        self.assertIsNone(res['mgmt_port'], res)
        self.assertEqual(1, len(res))

    def test_get_hosting_device_resources_by_complementary_id(self):
        tenant_id = 'some_tenant_id'
        ctx = bc.context.Context('', tenant_id, is_admin=True)
        mgmt_context = {'mgmt_nw_id': None}
        res = self.plugging_driver.create_hosting_device_resources(
            ctx, "some_id", tenant_id, mgmt_context, 1)
        # ports that should not be returned
        with self.port(), self.port(device_id='uuid2'), self.port(
                tenant_id=tenant_id):
            res_get = self.plugging_driver.get_hosting_device_resources(
                ctx, '', 'some_id', tenant_id, None)
            self.assertIsNone(res_get['mgmt_port'])
            self.assertEqual(1, len(res))

    def test_get_hosting_device_resources_by_device_id(self):
        tenant_id = 'some_tenant_id'
        ctx = bc.context.Context('', tenant_id, is_admin=True)
        mgmt_context = {'mgmt_nw_id': None}
        res = self.plugging_driver.create_hosting_device_resources(
            ctx, "some_id", tenant_id, mgmt_context, 1)
        # update attributes of created ports to fake what Nova updates
        hd_uuid = 'hd_uuid1'
        update_spec = {'port': {'device_id': hd_uuid,
                                'device_owner': 'nova'}}
        for hd_port in self._list('ports')['ports']:
            self._update('ports', hd_port['id'], update_spec)
        # ports that should not be returned
        with self.port(), self.port(device_id='uuid2'), self.port(
                tenant_id=tenant_id), self.port(tenant_id=tenant_id,
                                                device_owner='other_uuid'):
            res_get = self.plugging_driver.get_hosting_device_resources(
                ctx, hd_uuid, 'some_id', tenant_id, None)
            self.assertIsNone(res_get['mgmt_port'])
            self.assertEqual(1, len(res))

    def test_delete_hosting_device_resources(self):
        tenant_id = 'some_tenant_id'
        ctx = bc.context.Context('', tenant_id, is_admin=True)
        mgmt_context = {'mgmt_nw_id': None}
        res = self.plugging_driver.create_hosting_device_resources(
            ctx, "some_id", tenant_id, mgmt_context, 2)
        nets = self._list('networks')
        self.assertEqual(0, len(nets['networks']))
        subnets = self._list('subnets')
        self.assertEqual(0, len(subnets['subnets']))
        ports = self._list('ports')
        self.assertEqual(0, len(ports['ports']))
        # avoid passing the mgmt port twice in argument list
        mgmt_port = res['mgmt_port']
        del res['mgmt_port']
        self.plugging_driver.delete_hosting_device_resources(
            ctx, tenant_id, mgmt_port, **res)
        nets = self._list('networks')['networks']
        # no networks and subnets should remain
        self.assertEqual(0, len(nets))
        subnets = self._list('subnets')['subnets']
        self.assertEqual(0, len(subnets))
        ports = self._list('ports')
        self.assertEqual(0, len(ports['ports']))

    def test_transit_nets_cfg_invalid_file_format(self):
        self.plugging_driver._cfg_file = fileutils.write_to_tempfile(
            ("""{
                'EDGENAT': {
                    'gateway_ip': '1.109.100.254',
                    'cidr_exposed': '1.109.100.1/24',
                    'segmentation_id': 1066
                }
             }
             {
                'EDGENATBackup': {
                    'gateway_ip': '1.209.200.254',
                    'cidr_exposed': '1.209.200.1/24',
                    'segmentation_id': 1066
                }
             }""").encode('utf-8')
        )
        # TODO(thbachman): couldn't get assertRaises to work here,
        # so used this construct instead
        try:
            # just accessing the member should trigger the exception
            self.plugging_driver.transit_nets_cfg
            self.assertTrue(False)
        except aci_vlan.AciDriverConfigInvalidFileFormat:
            self.assertTrue(True)
        fileutils.delete_if_exists(self.plugging_driver._cfg_file)

    def test_config_sanity_check(self):
        test_config1 = {
            'Datacenter-Out': {
                'cidr_exposed': '1.103.2.0/24'
            }
        }
        test_config2 = {
            'Datacenter-Out': {
                'gateway_ip': '1.103.2.1',
            }
        }
        test_config3 = {
            'Datacenter-Out': {
                'gateway_ip': '1.103.2.254',
                'cidr_exposed': '1.103.2.1/24',
            }
        }
        self.assertRaises(aci_vlan.AciDriverConfigMissingGatewayIp,
                          self.plugging_driver._sanity_check_config,
                          test_config1)
        self.assertRaises(aci_vlan.AciDriverConfigMissingCidrExposed,
                          self.plugging_driver._sanity_check_config,
                          test_config2)
        self.assertTrue(
            test_config3,
            self.plugging_driver._sanity_check_config(test_config3))

    def test_no_driver(self):
        self.plugging_driver._apic_driver = None
        self.core_plugin.mechanism_manager.mech_drivers = {}
        # TODO(thbachman): couldn't get assertRaises to work here,
        # so used this construct instead
        try:
            self.plugging_driver.apic_driver()
            self.assertTrue(False)
        except aci_vlan.AciDriverNoAciDriverInstalledOrConfigured:
            self.assertTrue(True)


class TestAciVLANTrunkingPlugDriverGbp(
    test_l3_router_appliance_plugin.L3RouterApplianceTestCaseBase,
    test_l3.L3NatTestCaseMixin):
    """GBP-specific workflow testing of ACI VLAN driver

    This tests the GBP-specific workflow for the ACI VLAN Trunking
    Plugging driver.
    """

    # we use router types defined in .ini file.
    configure_routertypes = False
    router_type = 'ASR1k_Neutron_router'

    def setUp(self):
        super(TestAciVLANTrunkingPlugDriverGbp, self).setUp(
            create_mgmt_nw=False)
        # save possible test_lib.test_config 'config_files' dict entry so we
        # can restore it after tests since we will change its value
        self._old_config_files = copy.copy(test_lib.test_config.get(
            'config_files'))
        # include config files for device manager service plugin and router
        # service plugin since we define a number of hosting device templates,
        # hosting devices and routertypes there
        self._add_device_manager_plugin_ini_file()
        self._add_router_plugin_ini_file()
        #TODO(bobmel): Fix bug in test_extensions.py and we can remove the
        # below call to setup_config()
        self.setup_config()
        self.mock_gbp_driver = mock.MagicMock()
        self.mock_gbp_plugin = mock.MagicMock()
        self.mock_gbp_plugin.policy_driver_manager.policy_drivers = {
            'apic': self.mock_gbp_driver}
        g_p_mock = mock.MagicMock()
        plugins = {'CORE': self.core_plugin,
                   'GROUP_POLICY': self.mock_gbp_plugin,
                   bc.constants.L3: self.l3_plugin,
                   cisco_constants.DEVICE_MANAGER: self.core_plugin}
        g_p_mock.side_effect = lambda svc='CORE': plugins.get(svc)
        mock.patch('networking_cisco.backwards_compatibility.get_plugin',
                   g_p_mock).start()
        plug = aci_vlan.AciVLANTrunkingPlugDriver()
        plug.apic_driver.gbp_plugin.get_l3p_id_from_router_id = mock.Mock(
            return_value='somerouterid')
        plug.apic_driver.l3out_vlan_alloc.get_vlan_allocated = self._stub_vlan
        self.plugging_driver = plug
        self.vlan_dict = {'net1': APIC_VLAN1,
                          'net2': APIC_VLAN2,
                          'Datacenter-Out': APIC_VLAN2}

    def tearDown(self):
        if self._old_config_files is None:
            test_lib.test_config.pop('config_files', None)
        else:
            test_lib.test_config['config_files'] = self._old_config_files

#        manager.NeutronManager.get_service_plugins = self._real_get_plugins
        super(TestAciVLANTrunkingPlugDriverGbp, self).tearDown()

    def _stub_vlan(self, net, vrf, vrf_tenant):
        return self.vlan_dict.get(net)

    def _gen_ext_net_name(self, name):
        return aci_vlan.APIC_OWNED + _uuid() + "-" + name

    def _set_apic_driver_mocks(self, router):
        apic_driver = self.plugging_driver.apic_driver
        apic_driver.gbp_plugin.get_l3p_id_from_router_id = mock.Mock(
            return_value=router['id'])
        apic_driver.get_vrf_details = mock.Mock(
            return_value={'l3_policy_id': router['id']})

    def _verify_vrf(self, vrf_id, router):
        self.assertEqual(router['id'], vrf_id)

    def test_extend_hosting_port_info_adds_segmentation_id_internal(self):
        with self.network(name=self._gen_ext_net_name(
                'Datacenter-Out')) as ext_network:
            ext_net = ext_network['network']
            ext_net_id = ext_net['id']
            self._set_net_external(ext_net_id)
            with self.subnet() as subnet1:
                sn1 = subnet1['subnet']
                gw_info = {'network_id': ext_net_id}
                with self.router(external_gateway_info=gw_info,
                                 tenant_id=sn1['tenant_id']) as router1:
                    r1 = router1['router']
                    hosting_info = {}
                    fake_port_db_obj = FakePortDb('fakeuuid',
                        sn1['network_id'],
                        bc.constants.DEVICE_OWNER_ROUTER_INTF, r1['id'])
                    fake_port_db_obj.hosting_info['segmentation_id'] = 50
                    hosting_device = {'id':
                        '00000000-0000-0000-0000-000000000002'}
                    tenant_id = 'tenant_uuid1'
                    ctx = bc.context.Context('', tenant_id, is_admin=True)
                    self._set_apic_driver_mocks(r1)
                    self.plugging_driver.extend_hosting_port_info(ctx,
                        fake_port_db_obj, hosting_device, hosting_info)
                    self.assertEqual('GigabitEthernet/1/0/1',
                                     hosting_info['physical_interface'])
                    self.assertEqual(50, hosting_info['segmentation_id'])
                    self.assertIsNone(hosting_info.get('vrf_id'))

    def test_extend_hosting_port_info_adds_segmentation_id_external(self):
        with self.network(name=self._gen_ext_net_name(
                'Datacenter-Out')) as ext_network:
            ext_net = ext_network['network']
            ext_net_id = ext_net['id']
            self._set_net_external(ext_net_id)
            with self.subnet(network=ext_network) as subnet1:
                sn1 = subnet1['subnet']
                hosting_info = {}
                gw_info = {'network_id': ext_net_id}
                with self.router(external_gateway_info=gw_info,
                                 tenant_id=sn1['tenant_id']) as router1:
                    r1 = router1['router']
                    fake_port_db_obj = FakePortDb('fakeuuid', ext_net_id,
                        bc.constants.DEVICE_OWNER_ROUTER_GW, r1['id'])
                    fake_port_db_obj.hosting_info['segmentation_id'] = 40
                    hosting_device = {'id':
                        '00000000-0000-0000-0000-000000000002'}
                    tenant_id = 'tenant_uuid1'
                    ctx = bc.context.Context('', tenant_id, is_admin=True)
                    self._set_apic_driver_mocks(r1)
                    self.plugging_driver.extend_hosting_port_info(ctx,
                        fake_port_db_obj, hosting_device, hosting_info)
                    self.assertEqual('GigabitEthernet/2/0/1',
                                     hosting_info['physical_interface'])
                    self.assertEqual(40, hosting_info['segmentation_id'])
                    self._verify_vrf(hosting_info['vrf_id'], r1)

    # Had to create this b/c the helper won't let you set the name
    def _create_subnet_with_name(self, net_id, cidr, name):
        data = {'subnet': {'network_id': net_id,
                           'cidr': cidr,
                           'name': name,
                           'ip_version': 4,
                           'tenant_id': self._tenant_id}}
        subnet_req = self.new_create_request('subnets', data, self.fmt)
        subnet_res = subnet_req.get_response(self.api)
        # Things can go wrong - raise HTTP exc with res code only
        # so it can be caught by unit tests
        if subnet_res.status_int >= webob.exc.HTTPClientError.code:
            raise webob.exc.HTTPClientError(code=subnet_res.status_int)
        return self.deserialize(self.fmt, subnet_res)

    def test_extend_hosting_port_info_adds_snat_subnets(self):
        TEST_NET_NAME = 'Datacenter-Out'
        FAKE_IP = '1.1.1.2'
        FAKE_GW = '1.1.1.1'
        self.plugging_driver.apic_driver.get_snat_ip_for_vrf = mock.Mock(
            return_value={'external_segment_name': TEST_NET_NAME,
                'host_snat_ip': FAKE_IP,
                'gateway_ip': FAKE_GW,
                'prefixlen': 24})
        with self.network(name=self._gen_ext_net_name(
                TEST_NET_NAME)) as network:
            net = network['network']
            subnet = self._create_subnet_with_name(net['id'],
                                                   '10.0.0.0/24',
                                                   aci_vlan.APIC_SNAT_SUBNET)
            sn1 = subnet['subnet']
            ext_net_id = sn1['network_id']
            self._set_net_external(ext_net_id)
            gw_info = {'network_id': ext_net_id}
            with self.router(external_gateway_info=gw_info,
                             tenant_id=sn1['tenant_id']) as router1:
                r1 = router1['router']
                hosting_info = {}
                fake_port_db_obj = FakePortDb('fakeuuid', sn1['network_id'],
                    bc.constants.DEVICE_OWNER_ROUTER_GW, r1['id'])
                fake_port_db_obj.hosting_info['segmentation_id'] = 40
                hosting_device = {'id': '00000000-0000-0000-0000-000000000002'}
                tenant_id = 'tenant_uuid1'
                ctx = bc.context.Context('', tenant_id, is_admin=True)
                self._set_apic_driver_mocks(r1)
                self.plugging_driver.extend_hosting_port_info(ctx,
                    fake_port_db_obj, hosting_device, hosting_info)
                self.assertEqual([{'id': r1['tenant_id'],
                                   'ip': FAKE_IP,
                                   'cidr': sn1['cidr']}],
                                 hosting_info['snat_subnets'])

    def test_extend_hosting_port_info_adds_interface_configuration(self):
        TEST_INFO_CONFIG_LIST = ['testinfo1', 'testinfo2', 'testinfo3']
        self.plugging_driver._default_ext_dict = {
            'gateway_ip': '1.103.2.1',
            'cidr_exposed': '1.103.2.0/24',
            'interface_config': TEST_INFO_CONFIG_LIST
        }
        with self.network(name=self._gen_ext_net_name(
                'Datacenter-Out')) as network1:
            with self.subnet(network=network1) as subnet1:
                sn1 = subnet1['subnet']
                ext_net_id = sn1['network_id']
                self._set_net_external(ext_net_id)
                gw_info = {'network_id': ext_net_id}
                with self.router(external_gateway_info=gw_info,
                                 tenant_id=sn1['tenant_id']) as router1:
                    r1 = router1['router']
                    hosting_info = {}
                    fake_port_db_obj = FakePortDb('fakeuuid',
                        sn1['network_id'],
                        bc.constants.DEVICE_OWNER_ROUTER_INTF,
                        r1['id'])
                    hosting_device = {'id':
                                      '00000000-0000-0000-0000-000000000002'}
                    tenant_id = 'tenant_uuid1'
                    ctx = bc.context.Context('', tenant_id, is_admin=True)
                    self._set_apic_driver_mocks(r1)
                    self.plugging_driver.extend_hosting_port_info(ctx,
                        fake_port_db_obj, hosting_device, hosting_info)

                    self.assertIsNotNone(hosting_info.get('interface_config'))
                    for config in hosting_info['interface_config']:
                        self.assertIn(config, TEST_INFO_CONFIG_LIST)

    def test_extend_hosting_port_info_adds_global_configuration(self):
        TEST_INFO_CONFIG_LIST = ['testinfo1', 'testinfo2', 'testinfo3']
        self.plugging_driver._default_ext_dict = {
            'gateway_ip': '1.103.2.1',
            'cidr_exposed': '1.103.2.0/24',
            'global_config': TEST_INFO_CONFIG_LIST
        }
        dummy_router = {'id': 'someuuid',
                        'tenant_id': 'sometenantid',
                        ROUTER_ROLE_ATTR: None}
        self.plugging_driver.l3_plugin.get_router = mock.Mock(
            return_value=dummy_router)
        with self.network(name=self._gen_ext_net_name(
                'Datacenter-Out')) as network1:
            with self.subnet(network=network1) as subnet1:
                sn1 = subnet1['subnet']
                ext_net_id = sn1['network_id']
                self._set_net_external(ext_net_id)
                hosting_info = {}
                fake_port_db_obj = FakePortDb('fakeuuid',
                    sn1['network_id'],
                    bc.constants.DEVICE_OWNER_ROUTER_GW,
                    dummy_router['id'])
                hosting_device = {'id':
                                  '00000000-0000-0000-0000-000000000002'}
                tenant_id = 'tenant_uuid1'
                ctx = bc.context.Context('', tenant_id, is_admin=True)
                self._set_apic_driver_mocks(dummy_router)
                self.plugging_driver.extend_hosting_port_info(ctx,
                    fake_port_db_obj, hosting_device, hosting_info)

                self.assertIsNotNone(hosting_info.get('global_config'))
                for config in hosting_info['global_config']:
                    self.assertIn(config, TEST_INFO_CONFIG_LIST)

    def _update_provider_net_info(self, res_list, fields):
        for res in res_list:
            pv_info = self._pv_info['vlan'].get(res['id'])
            if pv_info is None:
                pv_info = self._pv_info['vxlan'].get(res['id'])
                if pv_info is None:
                    nw_type = self._network_type
                    if not self._pv_info[nw_type]:
                        tag = {'vlan': 11, 'vxlan': 7000}[nw_type]
                        pv_info = {'nw_type': nw_type, 'tag': tag}
                        self._pv_info[nw_type][res['id']] = pv_info
            if pv_info is None:
                tag = max([i['tag']
                           for i in self._pv_info[nw_type].values()]) + 1
                pv_info = {'nw_type': nw_type, 'tag': tag}
                self._pv_info[nw_type][res['id']] = pv_info
            res[bc.provider_net.NETWORK_TYPE] = pv_info['nw_type']
            res[bc.provider_net.SEGMENTATION_ID] = pv_info['tag']
            if fields is not None:
                for attr in list(res):
                    if attr not in fields:
                        del res[attr]

    def _mocked_get_network(self, context, id, fields=None):
        res = self.real_get_network(context, id)
        self._update_provider_net_info([res], fields)
        return res

    def _mocked_get_networks(self, *args, **kwargs):
        if len(args) >= 3:
            fields = args[2]
            list_args = [i for i in args]
            list_args[2] = None
            args = list_args
        else:
            fields = kwargs.pop('fields', None)
        res_list = self.real_get_networks(*args, **kwargs)
        self._update_provider_net_info(res_list, fields)
        return res_list

    def _test_allocate_hosting_port(self, test_info1):

        def _validate_allocation(plugin, ctx, r, port_db, test_info,
                                 i, hd, plugging_driver):
            binding_db = plugin._allocate_hosting_port(
                ctx, r['id'], port_db, hd['id'], plugging_driver)
            self.assertIsNotNone(binding_db.hosting_port_id)
            self.assertEqual(port_db.id, binding_db.hosting_port_id)
            self.assertEqual(test_info['vlan_tags'][i],
                             binding_db.segmentation_id)
        with self.network(name=self._gen_ext_net_name(
                'Datacenter-Out')) as ext_network:
            ext_net = ext_network['network']
            ext_net_id = ext_net['id']
            self._set_net_external(ext_net_id)
            with self.subnet(network=ext_network) as subnet1:
                sn1 = subnet1['subnet']
                gw_info = {'network_id': ext_net_id}
                with self.router(external_gateway_info=gw_info,
                             tenant_id=sn1['tenant_id']) as router1:
                    r1 = router1['router']
                    hds = self._list('hosting_devices')['hosting_devices']
                    hd = hds[0]
                    self._pv_info = {'vlan': {}, 'vxlan': {}}
                    self._network_type = test_info1['network_types'][0]
                    self.real_get_network = self.core_plugin.get_network
                    self.real_get_networks = self.core_plugin.get_networks
                    self._set_apic_driver_mocks(r1)
                    with mock.patch.object(self.core_plugin,
                                           'get_network') as m1,\
                            mock.patch.object(self.core_plugin,
                                              'get_networks') as m2:
                        m1.side_effect = self._mocked_get_network
                        m2.side_effect = self._mocked_get_networks
                        u1_ctx = bc.context.Context('', r1['tenant_id'],
                                                 is_admin=True)
                        gw_port_db = self.core_plugin._get_ports_query(
                            u1_ctx, filters={'network_id': [ext_net_id]}).one()
                        _validate_allocation(
                            self.l3_plugin, u1_ctx, r1, gw_port_db,
                            test_info1, 0, hd, self.plugging_driver)
                        for i in range(1, len(test_info1['network_types'])):
                            cidr = '1.0.' + str(i) + '.0/24'
                            with self.subnet(cidr=cidr) as subnet2:
                                sn2 = subnet2['subnet']
                                itfc_info = self._router_interface_action(
                                    'add', r1['id'], sn2['id'], None)
                                self._network_type = test_info1[
                                    'network_types'][i]
                                port_db = self.core_plugin._get_port(
                                    u1_ctx, itfc_info['port_id'])
                                _validate_allocation(
                                    self.l3_plugin, u1_ctx, r1,
                                    port_db, test_info1,
                                    i, hd, self.plugging_driver)

    def test_allocate_hosting_port_vlan_network_all_unused(self):
        self._test_allocate_hosting_port({'network_types': ['vlan'],
                                          'vlan_tags': [APIC_VLAN1]})

    def test_allocate_hosting_port_vlan_network_vlan_already_allocated(self):
        self._test_allocate_hosting_port(
            {'network_types': ['vlan', 'vlan'],
             'vlan_tags': [APIC_VLAN1, APIC_VLAN2]})

    def test_allocate_hosting_port_vlan_network_not_found_failure(self):
        with self.network(name=self._gen_ext_net_name(
                'Datacenter-Out')) as ext_network:
            ext_net = ext_network['network']
            ext_net_id = ext_net['id']
            self._set_net_external(ext_net_id)
            with self.subnet() as subnet1:
                sn1 = subnet1['subnet']
                gw_info = {'network_id': ext_net_id}
                with self.router(external_gateway_info=gw_info,
                                 tenant_id=sn1['tenant_id']) as router1:
                    r1 = router1['router']
                    u_ctx = bc.context.Context(
                        '', r1['tenant_id'], is_admin=True)
                    gw_port_db = self.core_plugin._get_ports_query(
                            u_ctx, filters={'network_id': [ext_net_id]}).one()
                    self._set_apic_driver_mocks(r1)
                    allocations = self.plugging_driver.allocate_hosting_port(
                        u_ctx, r1['id'], gw_port_db,
                        'vlan', 'non_existant_uuid')
                    self.assertIsNone(allocations)

    def test_allocate_hosting_port_info_adds_segment_id(self):
        self.plugging_driver._default_ext_dict = {
            'gateway_ip': '1.103.2.254',
            'cidr_exposed': '1.103.2.1/24',
            'interface_config': 'testinfo1',
            'segmentation_id': 3003
        }
        with self.network(name=self._gen_ext_net_name(
                'Datacenter-Out')) as network1:
            net1 = network1['network']
            self._set_net_external(net1['id'])
            net1['provider:network_type'] = 'opflex'

            def _return_mocked_net(self, args):
                return net1

            with self.subnet(network=network1) as subnet1:
                sn1 = subnet1['subnet']
                fake_port_db_obj = FakePortDb(
                    'some_dummy_id',
                    sn1['network_id'],
                    bc.constants.DEVICE_OWNER_ROUTER_GW,
                    'dummy_id'
                )
                hosting_device = {'id': '00000000-0000-0000-0000-000000000002'}
                tenant_id = 'tenant_uuid1'
                dummy_rid = 'dummy_router_id'
                ctx = bc.context.Context('', tenant_id, is_admin=True)
                with mock.patch.object(self.core_plugin, 'get_network') as m1:
                    m1.side_effect = _return_mocked_net
                    allocations = self.plugging_driver.allocate_hosting_port(
                        ctx, dummy_rid, fake_port_db_obj,
                        'opflex', hosting_device['id'])
                    self.assertEqual(3003, allocations['allocated_vlan'])

    def test_allocate_hosting_port_info_exception(self):
        self.plugging_driver._default_ext_dict = {
            'gateway_ip': '1.103.2.254',
            'cidr_exposed': '1.103.2.1/24',
            'interface_config': 'testinfo1',
        }
        with self.network(name=self._gen_ext_net_name(
                'Datacenter-Out')) as network1:
            net1 = network1['network']
            self._set_net_external(net1['id'])
            net1['provider:network_type'] = 'opflex'

            def _return_mocked_net(self, args):
                return net1

            with self.subnet(network=network1) as subnet1:
                sn1 = subnet1['subnet']
                fake_port_db_obj = FakePortDb(
                    'some_dummy_id',
                    sn1['network_id'],
                    bc.constants.DEVICE_OWNER_ROUTER_GW,
                    'dummy_id'
                )
                hosting_device = {'id': '00000000-0000-0000-0000-000000000002'}
                tenant_id = 'tenant_uuid1'
                dummy_rid = 'dummy_router_id'
                ctx = bc.context.Context('', tenant_id, is_admin=True)
                with mock.patch.object(self.core_plugin, 'get_network') as m1:
                    m1.side_effect = _return_mocked_net
                    self.assertRaises(
                        aci_vlan.AciDriverConfigMissingSegmentationId,
                        self.plugging_driver.allocate_hosting_port,
                        ctx, dummy_rid, fake_port_db_obj,
                        'opflex', hosting_device['id'])


class TestAciVLANTrunkingPlugDriverNeutron(TestAciVLANTrunkingPlugDriverGbp):
    """Neutron-specific workflow testing of ACI VLAN driver

    This tests the Neutron-specific workflow for the ACI VLAN Trunking
    Plugging driver.
    """

    # we use router types defined in .ini file.
    configure_routertypes = False
    router_type = 'ASR1k_Neutron_router'

    def setUp(self):
        super(TestAciVLANTrunkingPlugDriverGbp, self).setUp(
            create_mgmt_nw=False)
        # save possible test_lib.test_config 'config_files' dict entry so we
        # can restore it after tests since we will change its value
        self._old_config_files = copy.copy(test_lib.test_config.get(
            'config_files'))
        # include config files for device manager service plugin and router
        # service plugin since we define a number of hosting device templates,
        # hosting devices and routertypes there
        self._add_device_manager_plugin_ini_file()
        self._add_router_plugin_ini_file()
        #TODO(bobmel): Fix bug in test_extensions.py and we can remove the
        # below call to setup_config()
        self.setup_config()
        self.l3_plugin._core_plugin.mechanism_manager = mock.MagicMock()
        plug = aci_vlan.AciVLANTrunkingPlugDriver()
        plug.apic_driver.l3out_vlan_alloc.get_vlan_allocated = self._stub_vlan
        plug.apic_driver.per_tenant_context = True
        self.plugging_driver = plug
        self.vlan_dict = {'net1': APIC_VLAN1,
                          'net2': APIC_VLAN2,
                          'Datacenter-Out': APIC_VLAN2}

    def tearDown(self):
        if self._old_config_files is None:
            test_lib.test_config.pop('config_files', None)
        else:
            test_lib.test_config['config_files'] = self._old_config_files

        super(TestAciVLANTrunkingPlugDriverGbp, self).tearDown()

    def _gen_ext_net_name(self, name):
        return name

    def _set_apic_driver_mocks(self, router):
        apic_driver = self.plugging_driver.apic_driver
        apic_driver.get_router_vrf_and_tenant = mock.Mock(
            return_value={'aci_name': router['id'],
                          'aci_tenant': router['tenant_id']})

    def _verify_vrf(self, vrf_id, router):
        if self.plugging_driver.apic_driver.per_tenant_context:
            self.assertEqual(router['tenant_id'], vrf_id)
        else:
            self.assertEqual(router['id'], vrf_id)

    def test_extend_hosting_port_info_adds_snat_subnets(self):
        TEST_NET_NAME = 'Datacenter-Out'
        FAKE_IP = '1.1.1.2'
        FAKE_GW = '1.1.1.1'
        self.plugging_driver.apic_driver.get_snat_ip_for_vrf = mock.Mock(
            return_value={'external_segment_name': TEST_NET_NAME,
                'host_snat_ip': FAKE_IP,
                'gateway_ip': FAKE_GW,
                'prefixlen': 24})
        with self.network(name=self._gen_ext_net_name(
                TEST_NET_NAME)) as network:
            ext_net = network['network']
            ext_net_id = ext_net['id']
            self._set_net_external(ext_net_id)
            with self.network(name=(aci_vlan.APIC_SNAT_NET + '-' +
                                    ext_net_id)) as snat_net:
                net = snat_net['network']
                subnet = self._create_subnet_with_name(
                    net['id'], '10.0.0.0/24', aci_vlan.APIC_SNAT_SUBNET)
                sn1 = subnet['subnet']
                gw_info = {'network_id': ext_net_id}
                with self.router(external_gateway_info=gw_info,
                                 tenant_id=sn1['tenant_id']) as router1:
                    r1 = router1['router']
                    hosting_info = {}
                    fake_port_db_obj = FakePortDb('fakeuuid', ext_net_id,
                        bc.constants.DEVICE_OWNER_ROUTER_GW, r1['id'])
                    fake_port_db_obj.hosting_info['segmentation_id'] = 40
                    hosting_device = {'id':
                                      '00000000-0000-0000-0000-000000000002'}
                    tenant_id = 'tenant_uuid1'
                    ctx = bc.context.Context('', tenant_id, is_admin=True)
                    self._set_apic_driver_mocks(r1)
                    self.plugging_driver.extend_hosting_port_info(ctx,
                        fake_port_db_obj, hosting_device, hosting_info)
                    self.assertEqual([{'id': r1['tenant_id'],
                                       'ip': FAKE_IP,
                                       'cidr': sn1['cidr']}],
                                     hosting_info['snat_subnets'])

    def test_extend_hosting_port_info_no_snat_subnets_1(self):
        TEST_NET_NAME = 'Datacenter-Out'
        FAKE_IP = '1.1.1.2'
        FAKE_GW = '1.1.1.1'
        self.plugging_driver.apic_driver.get_snat_ip_for_vrf = mock.Mock(
            return_value={'external_segment_name': TEST_NET_NAME,
                'host_snat_ip': FAKE_IP,
                'gateway_ip': FAKE_GW,
                'prefixlen': 24})
        with self.network(name=self._gen_ext_net_name(
                TEST_NET_NAME)) as network:
            ext_net = network['network']
            ext_net_id = ext_net['id']
            self._set_net_external(ext_net_id)
            with self.network() as snat_net:
                net = snat_net['network']
                subnet = self._create_subnet_with_name(
                    net['id'], '10.0.0.0/24', aci_vlan.APIC_SNAT_SUBNET)
                sn1 = subnet['subnet']
                gw_info = {'network_id': ext_net_id}
                with self.router(external_gateway_info=gw_info,
                                 tenant_id=sn1['tenant_id']) as router1:
                    r1 = router1['router']
                    hosting_info = {}
                    fake_port_db_obj = FakePortDb('fakeuuid', ext_net_id,
                        bc.constants.DEVICE_OWNER_ROUTER_GW, r1['id'])
                    fake_port_db_obj.hosting_info['segmentation_id'] = 40
                    hosting_device = {'id':
                                      '00000000-0000-0000-0000-000000000002'}
                    tenant_id = 'tenant_uuid1'
                    ctx = bc.context.Context('', tenant_id, is_admin=True)
                    self._set_apic_driver_mocks(r1)
                    self.plugging_driver.extend_hosting_port_info(ctx,
                        fake_port_db_obj, hosting_device, hosting_info)
                    self.assertEqual([], hosting_info['snat_subnets'])

    def test_extend_hosting_port_info_no_snat_subnets_2(self):
        TEST_NET_NAME = 'Datacenter-Out'
        FAKE_IP = '1.1.1.2'
        FAKE_GW = '1.1.1.1'
        self.plugging_driver.apic_driver.get_snat_ip_for_vrf = mock.Mock(
            return_value={'external_segment_name': TEST_NET_NAME,
                'host_snat_ip': FAKE_IP,
                'gateway_ip': FAKE_GW,
                'prefixlen': 24})
        with self.network(name=self._gen_ext_net_name(
                TEST_NET_NAME)) as ext_network:
            ext_net = ext_network['network']
            ext_net_id = ext_net['id']
            self._set_net_external(ext_net_id)
            with self.subnet(network=ext_network) as subnet1:
                sn1 = subnet1['subnet']
                gw_info = {'network_id': ext_net_id}
                with self.router(external_gateway_info=gw_info,
                                 tenant_id=sn1['tenant_id']) as router1:
                    r1 = router1['router']
                    hosting_info = {}
                    fake_port_db_obj = FakePortDb('fakeuuid', ext_net_id,
                        bc.constants.DEVICE_OWNER_ROUTER_GW, r1['id'])
                    fake_port_db_obj.hosting_info['segmentation_id'] = 40
                    hosting_device = {'id':
                        '00000000-0000-0000-0000-000000000002'}
                    tenant_id = 'tenant_uuid1'
                    ctx = bc.context.Context('', tenant_id, is_admin=True)
                    self._set_apic_driver_mocks(r1)
                    self.plugging_driver.extend_hosting_port_info(ctx,
                        fake_port_db_obj, hosting_device, hosting_info)
                    self.assertIsNone(hosting_info.get('snat_subnets'))

    def test_extend_hosting_port_adds_segmentation_id_external_1_vrf(self):
        self.plugging_driver.apic_driver.per_tenant_context = False
        with self.network(name=self._gen_ext_net_name(
                'Datacenter-Out')) as ext_network:
            ext_net = ext_network['network']
            ext_net_id = ext_net['id']
            self._set_net_external(ext_net_id)
            with self.subnet(network=ext_network) as subnet1:
                sn1 = subnet1['subnet']
                hosting_info = {}
                gw_info = {'network_id': ext_net_id}
                with self.router(external_gateway_info=gw_info,
                                 tenant_id=sn1['tenant_id']) as router1:
                    r1 = router1['router']
                    fake_port_db_obj = FakePortDb('fakeuuid', ext_net_id,
                        bc.constants.DEVICE_OWNER_ROUTER_GW, r1['id'])
                    fake_port_db_obj.hosting_info['segmentation_id'] = 40
                    hosting_device = {'id':
                        '00000000-0000-0000-0000-000000000002'}
                    tenant_id = 'tenant_uuid1'
                    ctx = bc.context.Context('', tenant_id, is_admin=True)
                    self._set_apic_driver_mocks(r1)
                    self.plugging_driver.extend_hosting_port_info(ctx,
                        fake_port_db_obj, hosting_device, hosting_info)
                    self.assertEqual('GigabitEthernet/2/0/1',
                                     hosting_info['physical_interface'])
                    self.assertEqual(40, hosting_info['segmentation_id'])
                    self._verify_vrf(hosting_info['vrf_id'], r1)

    def test_external_net_name(self):
        self.assertIsNotNone(self.plugging_driver.get_ext_net_name)

    def test_external_net_no_gw(self):
        class DummyPort(object):

            def __init__(self, router_id):
                self.device_id = router_id
                self.device_owner = None

        drv = self.plugging_driver
        tenant_id = 'some_tenant_id'
        ctx = bc.context.Context('', tenant_id, is_admin=True)
        with self.router() as router1:
            r1 = router1['router']
            dummy_port = DummyPort(r1['id'])
            net_dict, net = drv._get_external_network_dict(ctx, dummy_port)
            self.assertIsNone(net)
            self.assertEqual({}, net_dict)

    def test_allocate_hosting_port_no_router(self):
        drv = self.plugging_driver
        tenant_id = 'some_tenant_id'
        ctx = bc.context.Context('', tenant_id, is_admin=True)
        with self.port() as port1:
            p1 = port1['port']
            self.assertIsNone(drv.allocate_hosting_port(ctx,
                None, p1, None, None))

    def test_allocate_hosting_port_router_no_gw(self):
        drv = self.plugging_driver
        tenant_id = 'some_tenant_id'
        ctx = bc.context.Context('', tenant_id, is_admin=True)
        with self.port() as port1:
            p1 = port1['port']
            with self.router() as router1:
                r1 = router1['router']
                p1['device_owner'] = bc.constants.DEVICE_OWNER_ROUTER_INTF
                self.assertIsNone(drv.allocate_hosting_port(ctx,
                    r1['id'], p1, None, None))



# Copyright 2015 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


"""Exceptions used by DFA enabler"""

from neutron_lib import exceptions
from neutronclient.common import exceptions as ncexc

from networking_cisco._i18n import _


# Neutronclient exceptions
ConnectionFailed = ncexc.ConnectionFailed


class NetworkNotFound(exceptions.NotFound):

    """Network cannot be found."""

    message = _("Network %(network_id)s could not be found.")


class ConfigProfileNotFound(exceptions.NotFound):

    """Config Profile cannot be found."""

    message = _("Config profile for network %(network_id)s could "
                "not be found.")


class ConfigProfileFwdModeNotFound(exceptions.NotFound):

    """Config Profile forwarding mode cannot be found."""

    message = _("Forwarding Mode for network %(network_id)s "
                "could not be found.")


class ConfigProfileIdNotFound(exceptions.NotFound):

    """Config Profile ID cannot be found."""

    message = _("Config Profile %(profile_id)s could not be found.")


class ConfigProfileNameNotFound(exceptions.NotFound):

    """Config Profile name cannot be found."""

    message = _("Config Profile %(name)s could not be found.")


class ProjectIdNotFound(exceptions.NotFound):

    """Project ID cannot be found."""

    message = _("Project ID %(project_id)s could not be found.")


class DfaClientRequestFailed(exceptions.ServiceUnavailable):

    """Request to DCNM failed."""

    message = _("Request to DCNM failed: %(reason)s.")


class DfaAgentFailed(exceptions.ServiceUnavailable):

    """Failure in running DfaAgent."""

    message = _("OpenStack is not running: %(reason)s.")


class InvalidInput(exceptions.InvalidInput):

    """Invalid Input specified."""

    message = _("Invalid input for operation: %(error_message)s.")
